from pysisyphus.calculators.Gaussian16 import Gaussian16


class Gaussian09(Gaussian16):

    conf_key = "gaussian09"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fn_base = "gaussian09"
        self.inp_fn = f"{self.fn_base}.com"
        self.out_fn = f"{self.fn_base}.log"
        self.chk_fn = f"{self.fn_base}.chk"
        self.base_cmd = self.get_cmd()
        self.formchk_cmd = self.get_cmd("formchk")

    def __str__(self):
        return "Gaussian09 calculator"



import logging
import os
from pathlib import Path
import sys

import numpy as np

from pysisyphus.optimizers.closures import bfgs_multiply
from pysisyphus.optimizers.hessian_updates import double_damp
from pysisyphus.optimizers.poly_fit import poly_line_search


class MicroOptimizer:
    def __init__(
        self,
        geom,
        step="lbfgs",
        line_search=True,
        max_cycles=100_000_000,
        max_step=0.2,
        keep_last=10,
        rms_force=None,
        double_damp=True,
        dump=False,
        **kwargs,
    ):
        self.geometry = geom
        self.step_funcs = {
            "sd": self.sd_step,
            "cg": self.cg_step,
            "lbfgs": self.lbfgs_step,
        }
        self.step_func = self.step_funcs[step]
        self.line_search = line_search
        self.max_cycles = max_cycles
        self.max_step = max_step
        self.keep_last = keep_last
        self.double_damp = double_damp
        self.rms_force = rms_force
        self.dump = dump

        self.prev_energy = None
        self.prev_step = None
        self.prev_forces = None
        self.coord_diffs = list()
        self.grad_diffs = list()

        self.trj_fn = "microopt_trj.xyz"
        if Path(self.trj_fn).exists():
            os.remove(self.trj_fn)

        self.logger = logging.getLogger("optimizer")

        if kwargs:
            msg = "Got unsupported keyword arguments: " + ", ".join(
                [f"{k}={v}" for k, v in kwargs.items()]
            )
            self.log(msg)
        self.is_converged = False

    def log(self, msg):
        self.logger.debug(msg)

    def run(self):
        self.geometry.reparametrize()

        for self.cur_cycle in range(self.max_cycles):
            if self.dump:
                with open(self.trj_fn, "a") as handle:
                    handle.write(self.geometry.as_xyz() + "\n")

            results = self.geometry.get_energy_and_forces_at(self.geometry.coords)
            forces = results["forces"]
            energy = results["energy"]
            rms = np.sqrt(np.mean(forces ** 2))
            print(f"{self.cur_cycle:03d} rms(f)={rms:.6f}")
            sys.stdout.flush()
            if self.rms_force and rms <= self.rms_force:
                print("Converged!")
                self.is_converged = True
                break

            self.take_step(energy, forces)

    def take_step(self, energy, forces, return_step=False):
        self.log(
            f"Cycle {self.cur_cycle:03d}, energy={energy:.6f} au, "
            f"norm(forces)={np.linalg.norm(forces):.6f}"
        )

        ip_gradient, ip_step = None, None
        if self.line_search and (self.cur_cycle > 0):
            ip_energy, ip_gradient, ip_step = poly_line_search(
                energy,
                self.prev_energy,
                -forces,
                -self.prev_forces,
                self.prev_step,
            )
        # Use the interpolated gradient for the step if interpolation succeeded
        if (ip_gradient is not None) and (ip_step is not None):
            forces = -ip_gradient
        # Keep the original gradient when the interpolation failed, but use
        # zero (interpolated) step.
        else:
            ip_step = np.zeros_like(forces)

        # Calculate actual step with potentially interpolated forces
        if self.cur_cycle == 0:
            step = self.sd_step(forces)
        else:
            step = self.step_func(forces)

        # Form full step. If we did not interpolate or interpolation failed ip_step will be zero.
        step = step + ip_step
        # Restrict absolute value of step vector entries
        step *= min(self.max_step / np.abs(step).max(), 1)

        self.geometry.reparametrize()

        self.prev_step = step
        self.prev_energy = energy
        self.prev_forces = forces
        if return_step:
            return step
        else:
            new_coords = self.geometry.coords + step
            self.geometry.coords = new_coords

    def optimize(self):
        if not hasattr(self, "cur_cycle"):
            self.cur_cycle = 0
        step = self.take_step(
            self.geometry.energy, self.geometry.forces, return_step=True
        )
        self.cur_cycle += 1
        return step

    def sd_step(self, forces):
        step = forces
        return step

    def cg_step(self, forces):
        beta = forces.dot(forces) / self.prev_forces.dot(self.prev_forces)
        step = forces + beta * self.prev_step
        return step

    def lbfgs_step(self, forces):
        y = self.prev_forces - forces
        s = self.prev_step

        if self.double_damp:
            s, y = double_damp(s, y, s_list=self.coord_diffs, y_list=self.grad_diffs)
        self.grad_diffs.append(y)
        self.coord_diffs.append(s)

        # Drop superfluous oldest vectors
        self.coord_diffs = self.coord_diffs[-self.keep_last :]
        self.grad_diffs = self.grad_diffs[-self.keep_last :]

        step = bfgs_multiply(
            self.coord_diffs,
            self.grad_diffs,
            forces,
            gamma_mult=False,
            logger=self.logger,
        )

        return step



import pyparsing as pp

from pysisyphus.helpers_pure import file_or_str


@file_or_str(".psf")
def parse_psf(text):
    int_ = pp.common.integer
    real = pp.common.sci_real
    psf_style_comment = pp.Regex(r"\*.*")

    psf = pp.CaselessLiteral("PSF")
    token = pp.ZeroOrMore(
        pp.CaselessLiteral("EXT")
        | pp.CaselessLiteral("CMAP")
        | pp.CaselessLiteral("XPLOR")
        | pp.CaselessLiteral("AUTOG")
    ).set_results_name("token")
    ntitle = int_ + pp.CaselessLiteral("!NTITLE")
    natom = int_.set_results_name("natom") + pp.CaselessLiteral("!NATOM")

    # Atom records
    atom_name = pp.Word(pp.alphanums)
    zero = pp.Literal("0")
    atom_record = pp.Group(
        int_.set_results_name("id")
        + pp.Word(pp.alphas).set_results_name("segment")
        + int_.set_results_name("resid")
        + pp.Word(pp.alphanums).set_results_name("resname")  # VAL, ALA, TIP3, etc.
        + atom_name.set_results_name("atom_name")
        + atom_name.set_results_name("atom_type")
        + real.set_results_name("charge")
        + real.set_results_name("mass")
        + zero
    )

    new_section = int_ + pp.Literal("!")

    def get_section(lhs, rhs=None):
        parser = int_.set_results_name("nterm") + pp.CaselessLiteral(lhs)

        if rhs is not None:
            parser += pp.Literal(":") + pp.CaselessLiteral(rhs)
        # Negative lookahead prevents matching 'nterm' of the next section
        parser += pp.Group(pp.ZeroOrMore(~new_section + int_)).set_results_name("inds")
        parser = pp.Group(parser)
        return parser

    nbond = get_section("!NBOND", "bonds").set_results_name("nbond")
    ntheta = get_section("!NTHETA", "angles").set_results_name("ntheta")
    nphi = get_section("!NPHI", "dihedrals").set_results_name("nphi")
    nimphi = get_section("!NIMPHI", "impropers").set_results_name("nimphi")
    ndon = get_section("!NDON", "donors").set_results_name("ndon")
    nacc = get_section("!NACC", "acceptors").set_results_name("nacc")
    nnb = get_section("!NNB").set_results_name("nnb")
    ncrterm = get_section("!NCRTERM", "cross-terms").set_results_name("ncrterm")

    parser = (
        psf
        + token
        + ntitle
        + natom
        + pp.OneOrMore(atom_record).set_results_name("atoms")
        + nbond  # bonds
        + ntheta  # bends
        + nphi  # dihedrals
        + nimphi  # impropers
        + ndon  # donors
        + nacc  # acceptors
        + nnb  # Non bonding something?
        # 0 0 !NGRP NST2
        # 0 0 !NUMLPH NUMLPH
        # + ncrterm
    )
    parser.ignore(psf_style_comment)

    result = parser.parseString(text)
    as_dict = result.asDict()
    return as_dict



import numpy as np 
import h5py

from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import eigval_to_wavenumber


def save_hessian(h5_fn, geom, cart_hessian=None, energy=None, mult=None):
    if cart_hessian is None:
        cart_hessian = geom.cart_hessian

    if energy is None:
        energy = geom.energy

    if mult is None:
        mult = geom.calculator.mult

    if len(geom.atoms) > 1:
        proj_hessian = geom.eckart_projection(geom.mass_weigh_hessian(cart_hessian))
    else:
        proj_hessian = cart_hessian
    eigvals, _ = np.linalg.eigh(proj_hessian)
    vibfreqs = eigval_to_wavenumber(eigvals)

    masses = geom.masses
    atoms = geom.atoms
    coords3d = geom.coords3d

    # with h5py.File(h5_fn, "w") as handle:
    #     handle.create_dataset("hessian", data=cart_hessian)
    #     handle.create_dataset("vibfreqs", data=vibfreqs)
    #     handle.create_dataset("masses", data=masses)
    #     handle.create_dataset("coords3d", data=coords3d)

    #     try:
    #         handle.attrs["atoms"] = [atom.lower() for atom in atoms]
    #     except OSError:
    #         handle.create_dataset("atoms", data=[atom.lower() for atom in atoms])
    #     handle.attrs["energy"] = energy
    #     handle.attrs["mult"] = mult


def save_third_deriv(h5_fn, geom, third_deriv_result, H_mw, H_proj):
    # with h5py.File(h5_fn, "w") as handle:
    #     for key, value in third_deriv_result._asdict().items():
    #         handle.create_dataset(key, data=value)

    #     handle.create_dataset("coords3d", data=geom.coords3d)
    #     handle.create_dataset("masses", data=geom.masses)
    #     handle.create_dataset("H_mw", data=H_mw)
    #     handle.create_dataset("H_proj", data=H_proj)
    #     try:
    #         handle.attrs["atoms"] = [atom.lower() for atom in atoms]
    #     except OSError:
    #         handle.create_dataset("atoms", data=[atom.lower() for atom in atoms])
    pass


def geom_from_hessian(h5_fn, **geom_kwargs):
    with h5py.File(h5_fn, "r") as handle:
        try:
            atoms = [atom.capitalize() for atom in handle.attrs["atoms"]]
        except (KeyError, TypeError):
            atoms = [atom.capitalize() for atom in handle["atoms"][:]]
        coords3d = handle["coords3d"][:]
        energy = handle.attrs["energy"]
        cart_hessian = handle["hessian"][:]

    geom = Geometry(atoms=atoms, coords=coords3d, **geom_kwargs)
    geom.cart_hessian = cart_hessian
    geom.energy = energy
    return geom



# [1] https://doi.org/10.1002/jcc.23481
#     Exploring transition state structures for intramolecular pathways
#     by the artificial force induced reaction method
#     Maeda, Morokuma et al, 2013
# [2] https://pubs.acs.org/doi/10.1021/ct200290m
#     Finding Reaction Pathways of Type A + B → X: Toward Systematic
#     Prediction of Reaction Mechanisms
#     Maeda, Morokuma, 2011

from dataclasses import dataclass
import itertools as it
import logging
from functools import reduce
import os
from pathlib import Path
from pprint import pformat
import shutil
import traceback
from typing import Callable, Dict, List, Tuple, Optional

import numpy as np
from numpy.typing import NDArray
from rmsd import kabsch_rmsd
from scipy.spatial.distance import pdist
from scipy.optimize import least_squares

from pysisyphus import logger as pysis_logger
from pysisyphus.calculators.AFIR import AFIR, CovRadiiSumZero
from pysisyphus.calculators import HardSphere
from pysisyphus.config import AFIR_RMSD_THRESH, OUT_DIR_DEFAULT
from pysisyphus.constants import BOHR2ANG
from pysisyphus.cos.NEB import NEB
from pysisyphus.drivers.opt import run_opt
from pysisyphus.elem_data import COVALENT_RADII as CR
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import pick_image_inds, check_for_end_sign
from pysisyphus.helpers_pure import to_sets
from pysisyphus.intcoords.helpers import get_bond_difference
from pysisyphus.intcoords.setup import get_pair_covalent_radii
from pysisyphus.intcoords.setup_fast import find_bonds
from pysisyphus.optimizers.FIRE import FIRE
from pysisyphus.xyzloader import make_xyz_str


logger = pysis_logger.getChild("afir")
logger.setLevel(logging.DEBUG)
file_handler = logging.FileHandler("afir.log", mode="w", delay=True)
logger.addHandler(file_handler)


AFIR_BOND_FACTOR = 1.2


@dataclass
class AFIRPath:
    atoms: tuple
    cart_coords: np.ndarray
    energies: np.ndarray
    forces: np.ndarray
    charge: int
    mult: int
    opt_is_converged: Optional[bool] = None
    gamma: Optional[float] = None
    path_indices: Optional[List[int]] = None

    def compatible(self, other):
        check = ("atoms", "charge", "mult")
        return all([getattr(self, name) == getattr(other, name) for name in check])

    def __add__(self, other):
        """This assumes that the first item in other.cart_coords is the same as the
        last item of self.cart_coords."""
        assert self.compatible(other)

        def conc(name):
            return np.concatenate(
                (getattr(self, name), getattr(other, name)[1:]), axis=0
            )

        cart_coords = conc("cart_coords")
        energies = conc("energies")
        forces = conc("forces")

        if self.path_indices is None:
            path_indices = [0] * len(self.cart_coords)
        else:
            path_indices = self.path_indices
        path_indices += [path_indices[-1] + 1] * (len(other.cart_coords) - 1)

        return AFIRPath(
            atoms=self.atoms,
            cart_coords=cart_coords,
            energies=energies,
            forces=forces,
            charge=self.charge,
            mult=self.mult,
            path_indices=path_indices,
        )

    def dump_trj(self, fn):
        geom = Geometry(self.atoms, self.cart_coords[0])
        xyzs = [geom.as_xyz(cart_coords=cc) for cc in self.cart_coords]
        with open(fn, "w") as handle:
            handle.write("\n".join(xyzs))


def analyze_afir_path(energies):
    energies = np.array(energies)
    max_ind = energies.argmax()

    local_minima = list()
    local_maxima = list()
    stationary_points = list()
    for i, en in enumerate(energies[1:-1], 1):
        prev_en = energies[i - 1]
        next_en = energies[i + 1]
        if is_minimum := prev_en > en < next_en:
            local_minima.append(i)
        if is_ts := prev_en < en > next_en:
            local_maxima.append(i)
        if is_minimum or is_ts:
            stationary_points.append(i)

    if max_ind in stationary_points:
        ts_ind = np.where(stationary_points == max_ind)[0][0]
        prev_min_ind = stationary_points[ts_ind - 1]
        next_min_ind = stationary_points[ts_ind + 1]
        sp_inds = [prev_min_ind, max_ind, next_min_ind]
    else:
        sp_inds = list()
    return sp_inds


##########################
#  Multi-component AFIR  #
#        Helper          #
##########################


def generate_random_union(geoms, offset=2.0, rng=None):
    """Unite fragments into one Geometry with random fragment orientations.

    Center, rotate and translate from origin acoording to approximate radius
    and an offset.
    Displace along +x, -x, +y, -y, +z, -z.

    Results for > 3 fragments don't look so pretty ;).
    """

    axis_inds = (0, 0, 1, 1, 2, 2)
    axis_inds_num = len(axis_inds)
    axis_translations = np.zeros(axis_inds_num)
    randomized = list()
    for i, geom in enumerate(geoms):
        i_mod = i % axis_inds_num
        geom = geom.copy()
        geom.center()
        geom.rotate(rng=rng)
        axis = axis_inds[i_mod]
        step_size = axis_translations[i_mod] + geom.approximate_radius() + offset
        axis_translations[i_mod] = step_size
        step = np.zeros(3)
        # Alternate between negative and positive direction along x/y/z
        step[axis] = (-1) ** i * step_size
        geom.coords3d += step[None, :]
        randomized.append(geom)
    union = reduce(lambda geom1, geom2: geom1 + geom2, randomized)
    return union


def generate_random_union_ref(geoms, rng=None, opt_kwargs=None):
    """Unite fragments into one Geometry with random fragment orientations."""

    geoms = [geom.copy() for geom in geoms]

    if rng is None:
        rng = np.random.default_rng()

    if opt_kwargs is None:
        opt_kwargs = {}

    # Random rotations
    for geom in geoms:
        geom.center()
        geom.rotate(rng=rng)

    # HardSphere optimization to fix overlapping fragments
    #
    # Set up fragment lists.
    fragments = list()
    for geom in geoms:
        geom_inds = np.arange(len(geom.atoms))
        try:
            geom_inds += fragments[-1][-1] + 1
        except IndexError:
            pass
        fragments.append(geom_inds.tolist())
    union = reduce(lambda geom1, geom2: geom1 + geom2, geoms)
    calc = HardSphere(union, frags=fragments, permutations=False, kappa=1.0)
    union.set_calculator(calc)
    _opt_kwargs = {
        "max_step": 0.2,
        "max_cycles": 500,
    }
    _opt_kwargs.update(opt_kwargs)
    opt = FIRE(union, **_opt_kwargs)
    opt.run()
    if not opt.is_converged:
        union = None
    else:
        # Remove HardSphere calculator
        union.clear()
        del union.calculator

    return union


def prepare_mc_afir(geoms, rng=None, **kwargs):
    """Wrapper for generate_random_union(_ref)."""
    union = generate_random_union(geoms, rng=rng, **kwargs)

    # Set up list of fragments
    i = 0
    fragments = list()
    for geom in geoms:
        atom_num = len(geom.atoms)
        fragments.append(np.arange(atom_num) + i)
        i += atom_num

    afir_kwargs = {
        "fragment_indices": fragments,
    }
    broken_bonds = []
    return union, afir_kwargs, broken_bonds


###########################
#  Single-component AFIR  #
#        Helper           #
###########################


def decrease_distance(coords3d, m, n, frac=0.8):
    c3d_m = coords3d[m]
    c3d_n = coords3d[n]
    dist_vec = c3d_n - c3d_m
    step = (1 - frac) / 2 * dist_vec
    c3d_new = coords3d.copy()
    c3d_new[m] += step
    c3d_new[n] -= step
    return c3d_new


def lstsqs_with_reference(coords3d, ref_coords3d, freeze_atoms=None):
    """Least-squares w.r.t. reference coordinates while keeping some
    atoms frozen."""

    if freeze_atoms is None:
        freeze_atoms = []
    else:
        freeze_atoms = list(freeze_atoms)
    ref_dists = pdist(ref_coords3d)

    mask = np.ones_like(coords3d[:, 0], dtype=bool)
    mask[freeze_atoms] = False
    # All atoms w/o the frozen atoms
    coords = coords3d[mask].flatten()
    x0 = coords

    coords_full = coords3d.copy()

    def fun(x):
        # Consider all distances, including distances to the fixed atoms 'm' and 'n'.
        coords_full[mask] = x.reshape(-1, 3)
        dists = pdist(coords_full)
        return dists - ref_dists

    res = least_squares(fun, x0)
    opt_coords = res.x
    coords_full[mask] = opt_coords.reshape(-1, 3)
    return res, coords_full


def weight_function(atoms, coords3d, i, j, p=6):
    cr_sum = sum([CR[atoms[k].lower()] for k in (i, j)])
    r_ij = np.linalg.norm(coords3d[i] - coords3d[j])
    omega = (cr_sum / r_ij) ** p
    return omega


def find_candidates(center, bond_sets):
    center_set = {
        center,
    }
    bonded_to_center = list()
    for bond in bond_sets:
        if center not in bond:
            continue
        bonded_to_center.append(*set(bond) - center_set)
    return bonded_to_center


def automatic_fragmentation(
    atoms, coords3d, frag1, frag2, cycles=2, p=6, bond_factor=1.25
):
    """Automatic fragmentation scheme as described in SC-AFIR paper [1]."""

    frag1 = set(frag1)
    frag2 = set(frag2)

    def w(m, n):
        """Shortcut for weight function"""
        return weight_function(atoms, coords3d, m, n, p=p)

    pairs = list(it.product(frag1, frag2))
    weights = [w(m, n) for m, n in pairs]
    max_weight = max(weights)

    bonds = find_bonds(atoms, coords3d, bond_factor=bond_factor)
    bond_sets = [set(bond) for bond in bonds.tolist()]

    def filter_candidates(candidates, partners, max_weight):
        to_keep = set()
        for candidate in candidates:
            for partner in partners:
                if candidate == partner:
                    break
                weight = w(candidate, partner)
                if weight > max_weight:
                    break
            else:
                to_keep.add(candidate)
        return to_keep

    def grow_fragment(frag1, frag2):
        # Find candidates that are bonded to atoms in frag1. Step 2 in [1].
        candidates = [find_candidates(m, bond_sets) for m in frag1]
        # Filter out candidates that are already contained in frag1
        candidates = [c for c in it.chain(*candidates) if c not in frag1]
        # Filter out candidates with weights that are too big. Step 3/4 in [1].
        candidates = set(filter_candidates(candidates, frag2, max_weight))
        return candidates

    for _ in range(cycles):
        f1_candidates = grow_fragment(frag1, frag2)
        f2_candidates = grow_fragment(frag2, frag1)

        # Step 5 in [1].
        f1_candidates = filter_candidates(f1_candidates, f2_candidates, max_weight)
        f2_candidates = filter_candidates(f2_candidates, f1_candidates, max_weight)

        # Step 6 in [1].
        frag1.update(f1_candidates)
        frag2.update(f2_candidates)
        assert frag1.isdisjoint(
            frag2
        ), "Overlapping fragments detected!"  # Sanity check
    return frag1, frag2


def prepare_sc_afir(geom, m, n, bond_factor=AFIR_BOND_FACTOR):
    """Create perturbed geometry, determine fragments and set AFIR calculator."""
    geom = geom.copy()
    atoms = geom.atoms
    org_coords3d = geom.coords3d.copy()

    def bond_sets(coords3d):
        bonds = find_bonds(atoms, coords3d, bond_factor=bond_factor)
        return set([frozenset(bond) for bond in bonds.tolist()])

    org_bond_sets = bond_sets(org_coords3d)

    # Move target atoms closer together along distance vector (decrease distance)
    decr_coords3d = decrease_distance(geom.coords3d, m, n)
    # Optimize remaining coordinates using least-squares, while keeping target
    # atom pair fixed.
    _, opt_coords3d = lstsqs_with_reference(decr_coords3d.copy(), geom.coords3d, (m, n))
    # Determine fragments, using the automated fragmentation
    frag1, frag2 = automatic_fragmentation(atoms, opt_coords3d, [m], [n])
    logger.debug(f"Fragments for target pair [{m}, {n}]: ({frag1}, {frag2})")
    fragment_indices = [list(frag) for frag in (frag1, frag2)]
    # Set lstsq-optimized coordinates and created wrapped calculator
    geom.coords3d = opt_coords3d

    opt_bond_sets = bond_sets(opt_coords3d)
    broken_bonds = org_bond_sets - opt_bond_sets

    afir_kwargs = {
        "fragment_indices": fragment_indices,
        # If 'complete_fragments' would be True, all remaining atom indices not
        # present in 'fragment_indices' would be assigned to a third fragment.
        "complete_fragments": False,
    }

    def set_atoms(inds, atom_type="X", mod_atoms=None):
        if mod_atoms is None:
            mod_atoms = list(atoms)
        for i in inds:
            mod_atoms[i] = atom_type
        return mod_atoms

    atoms_target = set_atoms((m, n))
    atoms_fragments = set_atoms(frag1)
    atoms_fragments = set_atoms(frag2, atom_type="Q", mod_atoms=atoms_fragments)

    atoms_coords3d = {
        "original": (atoms, org_coords3d),
        "original w/ target atoms": (atoms_target, org_coords3d),
        "decreased distance": (atoms, decr_coords3d),
        "decreased distance w/ target atoms": (atoms_target, decr_coords3d),
        "lstsq optimized": (atoms, opt_coords3d),
        "lstsq optimized w/ target atoms": (atoms_target, opt_coords3d),
        "lstsq optimized w/ fragments": (atoms_fragments, opt_coords3d),
    }
    trj = "\n".join(
        [
            make_xyz_str(atoms, BOHR2ANG * coords3d, comment=key)
            for key, (atoms, coords3d) in atoms_coords3d.items()
        ]
    )
    return geom, afir_kwargs, broken_bonds, trj


def determine_target_pairs(
    atoms: Tuple[str],
    coords3d: NDArray,
    min_: float = 1.25,
    max_: float = 5.0,
    active_atoms=None,
) -> List[Tuple[int]]:
    """Determine possible target m, n atom pairs for SC-AFIR calculations."""
    if active_atoms is None:
        active_atoms = range(len(atoms))
    active_atoms = set(active_atoms)

    pair_cov_radii = get_pair_covalent_radii(atoms)
    pair_dists = pdist(coords3d)
    quots = pair_dists / pair_cov_radii
    pair_inds = it.combinations(range(len(atoms)), 2)
    target_pairs = list()
    for pair_ind, quot in zip(pair_inds, quots):
        if (min_ <= quot <= max_) and (set(pair_ind) & active_atoms):
            target_pairs.append(pair_ind)
    return target_pairs


def determine_target_pairs_for_geom(geom: Geometry, **kwargs) -> List[Tuple[int]]:
    """Determine possible target m, n atom pairs for SC-AFIR calculations
    from geom."""
    target_pairs = determine_target_pairs(geom.atoms, geom.coords3d, **kwargs)
    return target_pairs


def coordinates_similar(
    test_coords3d: NDArray, ref_coords3d: List[NDArray], rmsd_thresh: float = 1e-2
) -> Tuple[bool, int]:
    # When the reference coordinates are an empty list.
    if len(ref_coords3d) == 0:
        return False, -1
    test_centered3d = test_coords3d - test_coords3d.mean(axis=0)[None, :]
    for i, rcoords3d in enumerate(ref_coords3d):
        ref_centered3d = rcoords3d - rcoords3d.mean(axis=0)[None, :]
        rmsd_ = kabsch_rmsd(test_centered3d, ref_centered3d)
        if rmsd_ <= rmsd_thresh:
            break
    else:
        return False, -1
    return True, i


def geom_similar(test_geom: Geometry, ref_geoms: List[Geometry], **kwargs) -> bool:
    return coordinates_similar(
        test_geom.coords3d, [geom.coords3d for geom in ref_geoms], **kwargs
    )


########################
#  Actual AFIR drivers #
########################


def opt_afir_path(geom, calc_getter, afir_kwargs, opt_kwargs=None, out_dir=None):
    """Minimize geometry with AFIR calculator."""
    if opt_kwargs is None:
        opt_kwargs = dict()
    if out_dir is None:
        out_dir = "."
    out_dir = Path(out_dir)

    actual_calc = calc_getter(out_dir=out_dir / OUT_DIR_DEFAULT)

    def afir_calc_getter():
        afir_calc = AFIR(actual_calc, out_dir=out_dir, **afir_kwargs)
        return afir_calc

    _opt_kwargs = {
        "dump": True,
        "out_dir": out_dir,
        "prefix": "afir",
        "max_cycles": 125,
        "overachieve_factor": 3,
        "hessian_update": "flowchart",
    }
    logger.debug(
        "\n".join(
            (
                "afir_kwargs:",
                "\t" + pformat(afir_kwargs),
                "opt_kwargs:",
                "\t" + pformat(_opt_kwargs),
            )
        )
    )
    _opt_kwargs.update(opt_kwargs)
    opt_result = run_opt(geom, afir_calc_getter, opt_key="rfo", opt_kwargs=_opt_kwargs)
    opt = opt_result.opt

    afir_path = AFIRPath(
        atoms=geom.atoms,
        cart_coords=np.array(opt.cart_coords),
        energies=np.array(opt.true_energies),
        forces=np.array(opt.true_forces),
        opt_is_converged=opt.is_converged,
        charge=actual_calc.charge,
        mult=actual_calc.mult,
        gamma=afir_kwargs["gamma"],
    )

    return afir_path


def run_afir_path(
    geom,
    calc_getter,
    out_dir,
    gamma_max,
    gamma_interval: Tuple[float, float],
    rng,
    ignore_bonds=None,
    bond_factor=AFIR_BOND_FACTOR,
    afir_kwargs=None,
    opt_kwargs=None,
):
    """Driver for AFIR minimizations with increasing gamma values."""
    if ignore_bonds is None:
        ignore_bonds = list()
    if afir_kwargs is None:
        afir_kwargs = dict()
    if opt_kwargs is None:
        opt_kwargs = {}

    if out_dir.exists():
        dir_contents = os.listdir(out_dir)
        for fn in dir_contents:
            fn = out_dir / fn
            try:
                os.remove(fn)
            except IsADirectoryError:
                shutil.rmtree(fn)
    else:
        os.mkdir(out_dir)

    # Decreasing the distance between two atoms in SC-AFIR may lead to broken
    # bonds for these two bonds, e.g., hydrogen atoms "that are left behind".
    # These bonds will be formed again when the AFIR function is minimized, but
    # we are not interested in these changes. So they can be ignored here.
    ignore_bonds = set([frozenset(bond) for bond in ignore_bonds])

    ref_calc = calc_getter()
    ref_energy = ref_calc.get_energy(geom.atoms, geom.cart_coords)["energy"]
    geom_backup = geom.copy()

    # By using (1.0, 1.0) as interval we can directly start at gamma_max, e.g.,
    # in SC-AFIR.
    gamma_low, gamma_high = gamma_interval
    assert gamma_high >= gamma_low
    gamma_spread = gamma_high - gamma_low
    gamma_0 = (gamma_low + (gamma_spread * rng.random(1)[0])) * gamma_max
    gamma_inrc = 0.1 * gamma_max

    afir_paths = list()
    best_afir_path = None
    lowest_barrier = None
    gamma = gamma_0

    # Minimize AFIR functions until gamma exceeds gamma_max.
    logger.info(f"New AFIR run with γ_max={gamma_max:.6f} au")
    while gamma <= gamma_max:
        gamma_ratio = gamma / gamma_max
        logger.info(f"AFIR run with γ={gamma:.6f} au, γ/γ_max={gamma_ratio: >6.2%}")
        _afir_kwargs = afir_kwargs.copy()
        _afir_kwargs["gamma"] = gamma
        try:
            afir_path = opt_afir_path(
                geom,
                calc_getter,
                afir_kwargs=_afir_kwargs,
                opt_kwargs=opt_kwargs,
                out_dir=out_dir,
            )
        # Can happen in SC-AFIR runs when fragments comprise only hydrogens.
        except CovRadiiSumZero:
            logger.warning("Sum of covalent radii is 0.0!")
            best_afir_path = None
            break
        except Exception:
            logger.error(f"Optimization crashed!\n{traceback.format_exc()}")
            best_afir_path = None
            break

        afir_paths.append(afir_path)
        true_energies = afir_path.energies

        # Check for changes in bond topology by comparing the result of the current
        # minimization to the initial geometry.
        formed, broken = get_bond_difference(geom_backup, geom, bond_factor=bond_factor)
        formed = to_sets(formed) - ignore_bonds
        broken = to_sets(broken) - ignore_bonds
        if formed or broken:
            max_energy = true_energies.max()
            barrier = max_energy - ref_energy

            # Always store the first path that leads to a change in bond topology.
            if lowest_barrier is None:
                best_afir_path = afir_path
                lowest_barrier = barrier
            # Break when a path with a lower or higher barrier is detected. This
            # should happen in the cycle after the first change in bond toplogy was
            # detected.
            if barrier < lowest_barrier:
                best_afir_path = afir_path
                lowest_barrier = barrier
                break
            elif barrier > lowest_barrier:
                break
        else:
            barrier = None

        # Update values for next cycle
        gamma += gamma_inrc

    # reduce(lambda ap1, ap2: ap1 + ap2, afir_paths).dump_trj("dumped_trj.xyz")

    # Construct TS guess from highest energy point along the AFIR path.
    if best_afir_path:
        guess_ind = best_afir_path.energies.argmax()
        guess_coords = best_afir_path.cart_coords[guess_ind]
        ts_guess = Geometry(geom.atoms, guess_coords)
        ts_guess.dump_xyz(out_dir / "ts_guess.xyz")
        afir_path_merged = reduce(lambda ap1, ap2: ap1 + ap2, afir_paths)
    else:
        ts_guess = None
        afir_path_merged = None
    return ts_guess, afir_path_merged


def relax_afir_path(atoms, cart_coords, calc_getter, images=15, out_dir=None):
    """Sample imagef from AFIR path and do COS relaxation."""
    image_inds = pick_image_inds(cart_coords, images=images)
    images = [Geometry(atoms, cart_coords[i]) for i in image_inds]

    # Relax last image
    opt_kwargs = {
        "dump": True,
        "prefix": "last",
        "out_dir": out_dir,
    }
    last_image = images[-1]
    last_image_backup = last_image.copy()
    run_opt(last_image, calc_getter, opt_key="rfo", opt_kwargs=opt_kwargs)
    _, broken = get_bond_difference(last_image, last_image_backup)
    if broken:
        return

    cos_kwargs = {}
    cos = NEB(images, **cos_kwargs)
    cos_opt_kwargs = {
        "align": True,
        "dump": True,
        "max_cycles": 30,
        "out_dir": out_dir,
    }
    run_opt(cos, calc_getter, opt_key="lbfgs", opt_kwargs=cos_opt_kwargs)


def run_mc_afir_paths(
    geoms: List,
    calc_getter: Callable,
    gamma_max: float,
    rng,
    N_max: int = 5,
    gamma_interval: Tuple[float, float] = (0.0, 1.0),
    afir_kwargs: Optional[Dict] = None,
    opt_kwargs: Optional[Dict] = None,
):
    if afir_kwargs is None:
        afir_kwargs = dict()

    N = 0
    N_0 = 0
    fmt = " >4d"
    while True:
        geom, _afir_kwargs, *_ = prepare_mc_afir(geoms, rng=rng)
        afir_kwargs = afir_kwargs.copy()
        afir_kwargs.update(_afir_kwargs)
        out_dir = Path(f"out_{N:03d}")

        ts_guess, afir_path = run_afir_path(
            geom,
            calc_getter,
            out_dir,
            gamma_max,
            gamma_interval,
            rng=rng,
            afir_kwargs=afir_kwargs,
            opt_kwargs=opt_kwargs,
        )
        ts_guess_is_new = yield N, ts_guess, afir_path
        if ts_guess_is_new:
            N_0 = N
        # Allow up to consecutive N_max failures
        if (N - N_0) > N_max:
            break
        logger.debug(f"{N_0=:{fmt}}, {N=:{fmt}}, {N-N_0=:{fmt}}, {N_max=:{fmt}}, ")
        N += 1


def run_sc_afir_paths(
    geom: Geometry,
    calc_getter: Callable,
    gamma_max: float,
    rng,
    N_max: int = 5,
    N_sample: int = 0,
    gamma_interval: Tuple[float, float] = (1.0, 1.0),  # Start directly with gamma_max
    afir_kwargs: Optional[Dict] = None,
    opt_kwargs: Optional[Dict] = None,
    target_pairs: Optional[List] = None,
):
    if target_pairs is None:
        target_pairs = determine_target_pairs_for_geom(geom)

    if afir_kwargs is None:
        afir_kwargs = dict()
    if opt_kwargs is None:
        opt_kwargs = dict()
    _opt_kwargs = opt_kwargs.copy()

    i = 0
    while len(target_pairs) > 0:
        m, n = target_pairs.pop(0)
        logger.info(f"Running SC-AFIR with target_pair ({m}, {n}).")
        # _afir_kwargs will contain the automatically determined fragments
        geom_mod, _afir_kwargs, broken_bonds, trj = prepare_sc_afir(geom, m, n)
        afir_kwargs = afir_kwargs.copy()
        afir_kwargs.update(_afir_kwargs)

        # _opt_kwargs.update({
        # "fragments": [[m, ], [n, ]],
        # "monitor_frag_dists": 5,
        # })

        out_dir = Path(f"out_{i:03d}")
        ts_guess, afir_path = run_afir_path(
            geom_mod,
            calc_getter,
            out_dir,
            gamma_max,
            gamma_interval,
            rng=rng,
            ignore_bonds=broken_bonds,
            afir_kwargs=afir_kwargs,
            opt_kwargs=_opt_kwargs,
        )
        ts_guess_is_new = yield i, ts_guess, afir_path
        i += 1
        # Here, TS optimization & IRC integration could take place, when the TS is new.


def run_afir_paths(
    afir_key,
    geoms,
    calc_getter,
    afir_kwargs=None,
    opt_kwargs=None,
    seed=None,
    N_sample=None,
    rmsd_thresh: float = AFIR_RMSD_THRESH,
    **kwargs,
):
    if seed is None:
        rng = np.random.default_rng()
        seed = rng.integers(1_000_000_000_000)
    logger.info(f"{seed=}")
    rng = np.random.default_rng(seed)

    if afir_key == "sc":
        assert (
            len(geoms) == 1
        ), f"Expected only 1 geometry for SC-AFIR, but got {len(geoms)}!."
        geoms = geoms[0]

    afir_funcs = {
        "mc": run_mc_afir_paths,
        "sc": run_sc_afir_paths,
    }
    afir_func = afir_funcs[afir_key]
    afir_coroutine = afir_func(
        geoms,
        calc_getter,
        rng=rng,
        afir_kwargs=afir_kwargs,
        opt_kwargs=opt_kwargs,
        **kwargs,
    )
    ts_guesses = list()
    afir_paths = list()
    stop_sign = "afir_stop"

    i, ts_guess, afir_path = next(afir_coroutine)
    while True:
        prefix = f"{i:03d}"
        logger.info(f"AFIR run {prefix}")
        # Check if we found enough TS guesses
        if N_sample and (len(ts_guesses) >= N_sample):
            break

        # Check similarity of TS guess to previously found TS guesses
        try:
            rmsds = [ts_guess.rmsd(guess) for guess in ts_guesses]
            min_rmsd = min(rmsds)
        except AttributeError:  # Raised when ts_guess is None
            min_rmsd = None
        except ValueError:
            min_rmsd = rmsd_thresh if (ts_guess is not None) else None

        if ts_guess_is_new := ((min_rmsd is not None) and (min_rmsd >= rmsd_thresh)):
            logger.debug(f"rmsds different enough! {min_rmsd=:.6f} au")
            ts_guesses.append(ts_guess)
            afir_paths.append(afir_path)
            afir_path.dump_trj(f"{prefix}_afir_path_trj.xyz")
            ts_guess.dump_xyz(f"{prefix}_ts_guess.xyz")
        elif min_rmsd:
            logger.debug(f"rmsds too similar! {min_rmsd=:.6f} au)")

        sign = check_for_end_sign(add_signs=(stop_sign,))
        if sign == stop_sign:
            break
        logger.info(f"{len(ts_guesses)=: >5d}")
        try:
            i, ts_guess, afir_path = afir_coroutine.send(ts_guess_is_new)
        except StopIteration:
            break

    return ts_guesses, afir_paths



# [1] https://aip.scitation.org/doi/abs/10.1063/1.3664901
#     Behn, 2011, Freezing string method
# [2] https://aip.scitation.org/doi/pdf/10.1063/1.4804162
#     Zimmerman, 2013, Growing string with interpolation and optimization
#                      in internal coordiantes

import numpy as np

from pysisyphus.optimizers.hessian_updates import double_damp
from pysisyphus.optimizers.closures import bfgs_multiply
from pysisyphus.optimizers.Optimizer import Optimizer
from pysisyphus.optimizers.restrict_step import scale_by_max_step


class StringOptimizer(Optimizer):
    def __init__(
        self,
        geometry,
        max_step=0.1,
        stop_in_when_full=-1,
        keep_last=10,
        lbfgs_when_full=True,
        gamma_mult=False,
        double_damp=True,
        scale_step="global",
        **kwargs,
    ):
        super().__init__(geometry, max_step=max_step, **kwargs)

        assert (
            self.is_cos
        ), "StringOptimizer is only intended to be used with COS objects."

        self.stop_in_when_full = int(stop_in_when_full)
        self.keep_last = int(keep_last)
        assert self.keep_last >= 0
        self.lbfgs_when_full = lbfgs_when_full
        if self.lbfgs_when_full and (self.keep_last == 0):
            print("lbfgs_when_full is True, but keep_last is 0!")
        self.gamma_mult = bool(gamma_mult)
        self.double_damp = bool(double_damp)
        self.scale_step = scale_step
        assert self.scale_step in ("global", "per_image")

        # Add one as we later subtract 1 before we check if this value is 0.
        self.stop_in = self.stop_in_when_full + 1
        self.is_cart_opt = self.geometry.coord_type == "cart"
        self.s_list = list()
        self.y_list = list()
        self.inds = list()

    def prepare_opt(self):
        if self.align and self.is_cart_opt:
            self.procrustes()

    def reset(self):
        pass

    def restrict_step_components(self, steps):
        too_big = np.abs(steps) > self.max_step
        self.log(f"Found {np.sum(too_big)} big step components.")
        signs = np.sign(steps[too_big])
        steps[too_big] = signs * self.max_step
        return steps

    def check_convergence(self, *args, **kwargs):
        # Normal convergence check with gradients etc.
        converged, conv_info = super().check_convergence(*args, **kwargs)

        if self.geometry.fully_grown:
            # We only start decrementing the counter after the string is fully grown.
            self.stop_in -= 1
            # Don't print this message if stop_in was disabled in the first place (< 0).
            if self.stop_in >= 0:
                self.log(f"String is fully grown. Stopping in {self.stop_in} cycle(s).")

        fully_grown = self.geometry.fully_grown
        full_stop = fully_grown and (self.stop_in == 0)
        # full_stop will take precedence when True.
        return full_stop or (fully_grown and converged), conv_info

    def optimize(self):
        new_image_inds = self.geometry.new_image_inds
        string_size_changed = len(new_image_inds) > 0

        if self.align and string_size_changed and self.is_cart_opt:
            self.procrustes()
            self.log("Aligned string.")

        forces = self.geometry.forces
        self.energies.append(self.geometry.energy)
        self.forces.append(forces)

        cur_size = self.geometry.string_size
        add_to_list = (
            self.keep_last > 0  # Only add to s_list and y_list if we want to keep
            and self.cur_cycle
            > 0  # cycles and if we can actually calculate differences.
            and (
                not self.lbfgs_when_full  # Add when LBFGS is allowed before fully grown.
                or self.lbfgs_when_full
                and self.geometry.fully_grown
                and not string_size_changed  # Don't add when string has to be fully grown
                # but grew fully in this cycle.
            )
        )
        if add_to_list:
            inds = list(range(cur_size))
            try:
                y = self.forces[-2] - forces
                s = self.coords[-1] - self.coords[-2]
            # Will be raised when the string grew in the previous cycle.
            except ValueError:
                cur_forces = np.delete(
                    forces.reshape(cur_size, -1), new_image_inds, axis=0
                ).flatten()
                y = self.forces[-2] - cur_forces
                cur_coords = np.delete(
                    self.coords[-1].reshape(cur_size, -1), new_image_inds, axis=0
                ).flatten()
                s = self.coords[-2] - cur_coords
                inds = np.delete(inds, new_image_inds)

            if self.double_damp:
                s, y = double_damp(s, y, s_list=self.s_list, y_list=self.y_list)

            self.s_list.append(s)
            self.y_list.append(y)
            self.inds.append(inds)
            # Drop oldest vectors
            self.s_list = self.s_list[-self.keep_last :]
            self.y_list = self.y_list[-self.keep_last :]
            self.inds = self.inds[-self.keep_last :]

        # Results in steepest descent step for empty s_list & y_list
        step = bfgs_multiply(
            self.s_list,
            self.y_list,
            forces,
            gamma_mult=self.gamma_mult,
            inds=self.inds,
            cur_size=cur_size,
            logger=self.logger,
        )

        # When LBFGS is not yet enabled then s_list and y_list will
        # be empty and the step from bfgs_multiply will be a simple SD step.
        # We try to calculated an improved step it via conjugate gradient.
        previous_step_with_same_size = (self.cur_cycle > 0) and (
            not string_size_changed
        )
        lbfgs_lists_empty = (len(self.s_list) == 0) and (len(self.y_list) == 0)
        if previous_step_with_same_size and lbfgs_lists_empty:
            prev_forces = self.forces[-2]
            # Fletcher-Reeves
            kind = "Fletcher-Reeves"
            beta = forces.dot(forces) / prev_forces.dot(prev_forces)
            # Polak-Ribiere
            # kind = "Polak-Ribiere"
            # beta = forces.dot(forces - prev_forces) / prev_forces.dot(prev_forces)
            beta = min(beta, 1)
            step = forces + beta * self.steps[-1]
            self.log(f"{kind} conjugate gradient correction, β={beta:.6f}")

        if self.scale_step == "global":
            step = scale_by_max_step(step, self.max_step)
        elif self.scale_step == "per_image":
            for image_step in step.reshape(len(self.geometry.images), -1):
                scale_by_max_step(image_step, self.max_step)
        else:
            raise Exception("Invalid scale_step={self.scale_step}!")

        return step



from pysisyphus.drivers.afir import run_afir_paths
from pysisyphus.drivers.diabatization import dq_diabatization_from_run_dict
from pysisyphus.drivers.opt import run_opt
from pysisyphus.drivers.scan import relaxed_scan, relaxed_1d_scan
from pysisyphus.drivers.birkholz import birkholz_interpolation
from pysisyphus.drivers.precon_pos_rot import run_precontr
from pysisyphus.drivers.perf import run_perf, print_perf_results
from pysisyphus.drivers.rates import (
    eyring_rate,
    harmonic_tst_rate,
    bell_corr,
    eckart_corr,
    eckart_corr_brown,
    wigner_corr,
)
from pysisyphus.drivers.replace import replace_atoms
from pysisyphus.drivers.spectrum import Spectrum



from collections import namedtuple
import io
import json
import os
import re
import shutil
import textwrap

import numpy as np

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.calculators.parser import parse_turbo_gradient
from pysisyphus.calculators.ORCA import save_orca_pc_file
from pysisyphus.constants import BOHR2ANG, BOHRPERFS2AU
from pysisyphus.helpers import geom_loader
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.xyzloader import make_xyz_str


OptResult = namedtuple("OptResult", "opt_geom opt_log")


class XTB(Calculator):

    conf_key = "xtb"
    _set_plans = (
        "charges",
        "json",
        "xtbrestart",
    )

    def __init__(
        self,
        gbsa="",
        alpb="",
        gfn=2,
        acc=1.0,
        iterations=250,
        etemp=None,
        retry_etemp=None,
        restart=False,
        topo=None,
        topo_update=None,
        quiet=False,
        **kwargs,
    ):
        """XTB calculator.

        Wrapper for running energy, gradient and Hessian calculations by
        XTB.

        Parameters
        ----------
        gbsa : str, optional
            Solvent for GBSA calculation, by default no solvent model is
            used.
        alpb : str, optional
            Solvent for ALPB calculation, by default no solvent model is
            used.
        gfn : int or str, must be (0, 1, 2, or "ff")
            Hamiltonian for the XTB calculation (GFN0, GFN1, GFN2, or GFNFF).
        acc : float, optional
            Accuracy control of the calculation, the lower the tighter several
            numerical thresholds are chosen.
        iterations : int, optional
            The number of iterations in SCC calculation.
        topo : str, optional
            Path the a GFNFF-topolgy file. As setting up the topology may take
            some time for sizable systems, it may be desired to reuse the file.
        topo_update : int
            Integer controlling the update interval of the GFNFF topology update.
            If supplied, the topolgy will be recreated every N-th calculation.
        mem : int
            Mememory per core in MB.
        quiet : bool, optional
            Suppress creation of log files.
        """
        super().__init__(**kwargs)

        self.gbsa = gbsa
        self.alpb = alpb
        self.gfn = gfn
        self.acc = acc
        self.iterations = iterations
        self.etemp = etemp
        self.retry_etemp = retry_etemp
        self.restart = restart
        if self.etemp is not None:
            assert (
                self.retry_etemp is None
            ), "Using 'etemp' and 'retry_etemp' simultaneously is not possible!"
        self.topo = topo
        self.topo_update = topo_update
        self.quiet = quiet

        self.topo_used = 0
        self.xtbrestart = None
        valid_gfns = (0, 1, 2, "ff")
        assert (
            self.gfn in valid_gfns
        ), f"Invalid gfn argument. Allowed arguments are: {', '.join(valid_gfns)}!"
        self.uhf = self.mult - 1

        self.inp_fn = "xtb.xyz"
        self.out_fn = "xtb.out"
        self.to_keep = (
            "out:xtb.out",
            "gradient",
            "xtbopt.xyz",
            "g98.out",
            "xtb_trj.xyz",
            # "json:xtbout.json",
            "charges:charges",
            "xcontrol",
        )
        if self.restart:
            self.to_keep += ("xtbrestart", )
        if self.quiet:
            self.to_keep = ()

        self.parser_funcs = {
            "grad": self.parse_gradient,
            "hess": self.parse_hessian,
            "opt": self.parse_opt,
            "md": self.parse_md,
            "topo": self.parse_topo,
            "noparse": lambda path: None,
            "calc": self.parse_energy,
        }

        self.base_cmd = self.get_cmd()

    def reattach(self, last_calc_cycle):
        pass

    def prepare_coords(self, atoms, coords):
        coords = coords * BOHR2ANG
        return make_xyz_str(atoms, coords.reshape((-1, 3)))

    def prepare_input(self, atoms, coords, calc_type, point_charges=None):
        path = self.prepare_path(use_in_run=True)

        xcontrol_str = """
        $write
            json=true
        $end
        """

        if point_charges is not None:
            pc_fn = self.make_fn("pointcharges_inp.pc")
            save_orca_pc_file(point_charges, pc_fn, hardness=99)
            xcontrol_str += f"""
            $embedding
               input={pc_fn}
               interface=orca
            $end
            """

        xcontrol_str = textwrap.dedent(xcontrol_str.strip())
        with open(path / "xcontrol", "w") as handle:
            handle.write(xcontrol_str)

        # Check if the topology has to be recreated/updated
        if (
            self.topo_used > 0
            and self.topo_update
            and (self.topo_used % self.topo_update == 0)
        ):
            results = self.run_topo(atoms, coords)
            self.topo = results["topo"]
            self.log(f"Updated topology! Saved to '{self.topo}'.")
        if self.topo:
            shutil.copy(self.topo, path / "gfnff_topo")
            self.log(f"Using toplogy given in {self.topo}.")
            self.topo_used += 1
        if self.xtbrestart is not None:
            shutil.copy(self.xtbrestart, path / "xtbrestart")
            self.log(f"Using xtbrestart given in {self.xtbrestart}.")

    def prepare_add_args(self, xcontrol=None):
        add_args = (
            f"--input xcontrol --chrg {self.charge} --uhf {self.uhf} "
            f"--acc {self.acc} --iterations {self.iterations}".split()
        )
        if self.etemp:
            etemp = f"--etemp {self.etemp}".split()
            add_args = add_args + etemp

        # Use solvent model if specified
        if self.gbsa:
            gbsa = f"--gbsa {self.gbsa}".split()
            add_args = add_args + gbsa
        elif self.alpb:
            alpb = f"--alpb {self.alpb}".split()
            add_args = add_args + alpb
        # Select parametrization
        gfn = ["--gfnff"] if self.gfn == "ff" else f"--gfn {self.gfn}".split()
        add_args = add_args + gfn
        return add_args

    def get_pal_env(self):
        env_copy = os.environ.copy()
        env_copy["OMP_NUM_THREADS"] = str(self.pal)
        env_copy["MKL_NUM_THREADS"] = str(self.pal)
        # Per thread
        env_copy["OMP_STACKSIZE"] = f"{self.mem}M"

        return env_copy

    def get_energy(self, atoms, coords, **prepare_kwargs):
        results = self.get_forces(atoms, coords, **prepare_kwargs)
        del results["forces"]
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        self.prepare_input(atoms, coords, "forces", **prepare_kwargs)
        inp = self.prepare_coords(atoms, coords)
        add_args = self.prepare_add_args() + ["--grad"]
        self.log(f"Executing {self.base_cmd} {add_args}")
        kwargs = {
            "calc": "grad",
            "add_args": add_args,
            "env": self.get_pal_env(),
        }
        results = self.run(inp, **kwargs)
        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        self.prepare_input(atoms, coords, "hessian", **prepare_kwargs)
        inp = self.prepare_coords(atoms, coords)
        add_args = self.prepare_add_args() + ["--hess"]
        self.log(f"Executing {self.base_cmd} {add_args}")
        kwargs = {
            "calc": "hess",
            "add_args": add_args,
            "env": self.get_pal_env(),
        }
        results = self.run(inp, **kwargs)
        return results

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        self.prepare_input(atoms, coords, "calculation", **prepare_kwargs)
        inp = self.prepare_coords(atoms, coords)
        kwargs = {
            "calc": "calc",
            "env": self.get_pal_env(),
        }
        energy = self.run(inp, **kwargs)
        results = {"energy": energy}
        return results

    def run_topo(self, atoms, coords):
        inp = self.prepare_coords(atoms, coords)
        kwargs = {
            "calc": "topo",
            "cmd": [self.base_cmd, "topo"],
            "env": self.get_pal_env(),
        }
        results = self.run(inp, **kwargs)
        return results

    def parse_topo(self, path):
        fn = "gfnff_topo"
        topo = path / fn
        target = self.make_fn(fn)
        shutil.copy(topo, target)

        return {
            "topo": target,
        }

    def get_mdrestart_str(self, coords, velocities):
        """coords and velocities have to given in au!"""
        vals = np.concatenate((coords, velocities), axis=1)

        with io.StringIO() as io_stream:
            np.savetxt(io_stream, vals, fmt="% .14e")
            mdrestart = io_stream.getvalue()
        # What does the -1.0 mean?
        mdrestart = "-1.0\n" + mdrestart.replace("e", "D")
        mdrestart = textwrap.indent(mdrestart, " ")
        return mdrestart

    def write_mdrestart(self, path, mdrestart_str):
        with open(path / "mdrestart", "wb") as handle:
            handle.write(mdrestart_str.encode("ascii"))

    def run_md(self, atoms, coords, t, dt, velocities=None, dump=1):
        """Expecting t and dt in fs, even though xtb wants t in ps!"""

        restart = "false"
        path = self.prepare_path(use_in_run=True)
        if velocities is not None:
            coords3d = coords.reshape(-1, 3)
            velocities3d = velocities.reshape(-1, 3) * BOHRPERFS2AU
            assert (
                coords3d.shape == velocities3d.shape
            ), "Shape of coordinates and velocities doesn't match!"
            mdrestart_str = self.get_mdrestart_str(coords3d, velocities3d)
            self.write_mdrestart(path, mdrestart_str)
            restart = "true"
        md_str = textwrap.dedent(
            """
        $md
            hmass=1
            dump={dump}  # fs
            nvt=false
            restart={restart}
            time={time}  # ps
            shake=0
            step={step}  # fs
            velo=false
        $end"""
        )
        t_ps = t / 1000
        md_str_fmt = md_str.format(restart=restart, time=t_ps, step=dt, dump=dump)
        with open(path / "xcontrol", "w") as handle:
            handle.write(md_str_fmt)
        inp = self.prepare_coords(atoms, coords)

        add_args = self.prepare_add_args() + ["--input", "xcontrol", "--md"]
        self.log(f"Executing {self.base_cmd} {add_args}")
        kwargs = {
            "calc": "md",
            "add_args": add_args,
            "env": self.get_pal_env(),
            "keep": True,
        }
        geoms = self.run(inp, **kwargs)
        return geoms

    def parse_md(self, path):
        assert (path / "xtbmdok").exists(), "File xtbmdok does not exist!"
        geoms = geom_loader(path / "xtb_trj.xyz")
        return geoms

    def run_opt(self, atoms, coords, keep=True, keep_log=False):
        inp = self.prepare_coords(atoms, coords)
        add_args = self.prepare_add_args() + ["--opt", "tight"]
        self.log(f"Executing {self.base_cmd} {add_args}")
        kwargs = {
            "calc": "opt",
            "add_args": add_args,
            "env": self.get_pal_env(),
            "keep": keep,
            "parser_kwargs": {"keep_log": keep_log},
        }
        opt_result = self.run(inp, **kwargs)
        return opt_result

    def parse_opt(self, path, keep_log=False):
        xtbopt = path / "xtbopt.xyz"
        if not xtbopt.exists():
            self.log(f"{self.calc_number:03d} failed")
            return None
        opt_geom = geom_loader(xtbopt)
        opt_geom.energy = self.parse_energy(path)

        opt_log = None
        if keep_log:
            opt_log = geom_loader(path / "xtbopt.log")

        opt_result = OptResult(opt_geom=opt_geom, opt_log=opt_log)
        return opt_result

    def parse_energy(self, path):
        with open(path / self.out_fn) as handle:
            text = handle.read()
        energy_re = r"TOTAL ENERGY\s*([-\d\.]+) Eh"
        energy = float(re.search(energy_re, text)[1])
        return energy

    def parse_gradient(self, path):
        return parse_turbo_gradient(path)

    def parse_hessian(self, path):
        with open(path / "hessian") as handle:
            text = handle.read()
        hessian = np.array(text.split()[1:], dtype=float)
        coord_num = int(hessian.size ** 0.5)
        hessian = hessian.reshape(coord_num, coord_num)
        energy = self.parse_energy(path)
        results = {
            "energy": energy,
            "hessian": hessian,
        }
        return results

    def parse_charges(self, fn=None):
        if fn is None:
            fn = self.charges
        charges = np.loadtxt(fn, dtype=float)
        return charges

    def parse_charges_from_json(self, fn=None):
        if fn is None:
            fn = self.json
        with open(fn, "r") as handle:
            dump = json.load(handle)
        charges = dump["partial charges"]
        return charges

    @staticmethod
    @file_or_str(".out")
    def check_termination(text):
        term_re = re.compile("finished run on")
        mobj = term_re.search(text)
        return bool(mobj)

    def get_retry_args(self):
        if self.retry_etemp is None:
            return []

        self.log(f"Retrying calculation with increased etemp={self.retry_etemp}")
        return f"--etemp {self.retry_etemp}".split()

    def __str__(self):
        return "XTB calculator"



import numpy as np

from pysisyphus.calculators.AnaPotBase import AnaPotBase


class Rastrigin(AnaPotBase):
    """
        http://www.sfu.ca/~ssurjano/rastr.html
    """

    def __init__(self): 
        # d == 2
        # f(x) = 10*d + sum_(i+1)^d (x_i**2 - 10*cos(2*pi*x_i))
        V_str = "20 + x**2 - 10*cos(2*pi*x) + y**2 - 10*cos(2*pi*y)"
        xlim = (-5.12, 5.12)
        ylim = xlim
        levels = np.linspace(0, 50, 50)
        minima = ((0., 0., 0.), )
        super().__init__(V_str=V_str, xlim=xlim, ylim=ylim, levels=levels, minima=minima)

    def __str__(self):
        return "Rastrigin calculator"



from pysisyphus.interpolate import IDPP, LST, Interpolator, Redund, Geodesic


INTERPOLATE = {
    "idpp": IDPP.IDPP,
    "lst": LST.LST,
    "redund": Redund.Redund,
    "linear": Interpolator.Interpolator,
    "geodesic": Geodesic.Geodesic,
}


def interpolate(initial_geom, final_geom, between, kind="linear",
                only_between=False, align=False, interpol_kwargs=None):
    if interpol_kwargs is None:
        interpol_kwargs = {}
    """ only_between: return only the interpolated images."""
    interpolate_class = INTERPOLATE[kind]
    interpolator = interpolate_class((initial_geom, final_geom), between,
                                     align=align, **interpol_kwargs)
    if only_between:
        return interpolator.interpolate(initial_geom, final_geom)
    else:
        return interpolator.interpolate_all()


def interpolate_all(geoms, between, kind="linear", align=False, **interpol_kwargs):
    if (between == 0) or (kind is None):
        return geoms

    interpolate_class = INTERPOLATE[kind]
    interpolator = interpolate_class(geoms, between, align=align, **interpol_kwargs)

    return interpolator.interpolate_all()



#!/usr/bin/env python3
"""Generate auto-derived CLI/YAML reference pages for docs."""

from __future__ import annotations

import argparse
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import click
import yaml
from click.testing import CliRunner

REPO_ROOT = Path(__file__).resolve().parents[1]
DOCS_ROOT = REPO_ROOT / "docs"
REF_ROOT = DOCS_ROOT / "reference"
COMMANDS_ROOT = REF_ROOT / "commands"
YAML_REF_PATH = REF_ROOT / "yaml.md"
TOOL_NAME = "pdb2reaction"

sys.path.insert(0, str(REPO_ROOT))

from pdb2reaction.cli import cli as root_cli  # noqa: E402


_ALL_TEMPLATE = """# Starter config for `pdb2reaction all`

extract:
  radius: 2.6
  radius_het2het: 0.0

calc:
  workers: 1
  workers_per_node: 1

path_search:
  mep_mode: gsm
  max_nodes: 20
  max_cycles: 300

scan:
  max_step_size: 0.2
  bias_k: 300.0
  relax_max_cycles: 10000

tsopt:
  max_cycles: 10000

freq:
  max_write: 10
  amplitude_ang: 0.8
  n_frames: 20
  sort: value
  temperature: 298.15
  pressure_atm: 1.0

dft:
  func_basis: wb97m-v/def2-tzvpd
  max_cycle: 100
  conv_tol: 1.0e-9
  grid_level: 3
"""


@dataclass(frozen=True)
class RenderedFile:
    path: Path
    content: str


@dataclass(frozen=True)
class CommandDoc:
    name: str
    stem: str


def _collect_subcommands() -> list[str]:
    ctx = click.Context(root_cli)
    return sorted(root_cli.list_commands(ctx))


def _doc_stem(command_name: str) -> str:
    # Keep command spelling in headings/tables, but standardize filenames.
    return command_name.replace("-", "_")


def _collect_command_docs() -> list[CommandDoc]:
    docs: list[CommandDoc] = []
    used_stems: dict[str, str] = {}
    for name in _collect_subcommands():
        stem = _doc_stem(name)
        prev = used_stems.get(stem)
        if prev is not None and prev != name:
            raise RuntimeError(
                f"Command doc stem collision: '{prev}' and '{name}' both map to '{stem}'."
            )
        used_stems[stem] = name
        docs.append(CommandDoc(name=name, stem=stem))
    return docs


def _capture_help(command_name: str, *, advanced: bool) -> str:
    runner = CliRunner()
    args = [command_name, "--help-advanced"] if advanced else [command_name, "--help"]
    result = runner.invoke(
        root_cli,
        args,
        catch_exceptions=False,
        prog_name=TOOL_NAME,
    )
    if result.exit_code != 0:
        raise RuntimeError(
            f"Failed to collect help for '{TOOL_NAME} {command_name}' "
            f"(advanced={advanced}):\n{result.output}"
        )
    # Strip version banner so docs stay stable across builds.
    lines = result.output.splitlines(keepends=True)
    lines = [ln for ln in lines if not ln.startswith(f"{TOOL_NAME} ver. ")]
    return "".join(lines).rstrip() + "\n"


def _render_command_page(command_name: str, help_text: str) -> str:
    return (
        f"# `{TOOL_NAME} {command_name}`\n\n"
        "```text\n"
        f"{help_text.rstrip()}\n"
        "```\n"
    )


def _render_commands_index(command_docs: list[CommandDoc]) -> str:
    toctree_body = "\n".join(doc.stem for doc in command_docs)
    rows = "\n".join(
        f"| `{TOOL_NAME} {doc.name}` | [{doc.name}]({doc.stem}.md) |" for doc in command_docs
    )
    return (
        "# CLI Command Reference\n\n"
        "```{toctree}\n"
        ":maxdepth: 1\n"
        ":hidden:\n\n"
        f"{toctree_body}\n"
        "```\n\n"
        "| Command | Page |\n"
        "|---|---|\n"
        f"{rows}\n"
    )


def _iter_scalar_rows(data: dict, prefix: str = "") -> Iterable[tuple[str, str, str]]:
    for key, value in data.items():
        full_key = f"{prefix}.{key}" if prefix else str(key)
        if isinstance(value, dict):
            yield from _iter_scalar_rows(value, prefix=full_key)
        else:
            type_name = type(value).__name__
            value_repr = repr(value)
            yield full_key, type_name, value_repr


def _render_yaml_reference() -> str:
    template_data = yaml.safe_load(_ALL_TEMPLATE) or {}
    top_keys = list(template_data.keys()) if isinstance(template_data, dict) else []
    top_rows = "\n".join(f"| `{k}` |" for k in top_keys)
    scalar_rows = "\n".join(
        f"| `{k}` | `{t}` | `{v}` |" for k, t, v in _iter_scalar_rows(template_data)
    )
    scan_schema = """```yaml
# scan (1D staged)
one_based: false
stages:
  - - [1, 2, 1.65]
  - - [2, 3, 2.30]

# scan2d / scan3d
one_based: false
pairs:
  - [1, 2, 1.40, 2.20]
  - [2, 3, 1.20, 2.00]
  - [3, 4, 1.00, 1.80]  # required only for scan3d
```"""

    return (
        "# YAML Schema\n\n"
        "## Top-level Keys\n\n"
        "| Key |\n"
        "|---|\n"
        f"{top_rows}\n\n"
        "## Starter Template\n\n"
        "```yaml\n"
        f"{_ALL_TEMPLATE.strip()}\n"
        "```\n\n"
        "## Scalar Defaults\n\n"
        "| Key | Type | Default |\n"
        "|---|---|---|\n"
        f"{scalar_rows}\n\n"
        "## Scan Spec Shapes\n\n"
        "Accepted by `scan`, `scan2d`, and `scan3d` with `-s/--scan-lists`.\n\n"
        f"{scan_schema}\n"
    )


def _render() -> list[RenderedFile]:
    command_docs = _collect_command_docs()
    rendered: list[RenderedFile] = []

    for doc in command_docs:
        try:
            help_text = _capture_help(doc.name, advanced=True)
        except RuntimeError:
            help_text = _capture_help(doc.name, advanced=False)
        rendered.append(
            RenderedFile(
                path=COMMANDS_ROOT / f"{doc.stem}.md",
                content=_render_command_page(doc.name, help_text),
            )
        )

    rendered.append(
        RenderedFile(
            path=COMMANDS_ROOT / "index.md",
            content=_render_commands_index(command_docs),
        )
    )
    rendered.append(RenderedFile(path=YAML_REF_PATH, content=_render_yaml_reference()))
    return rendered


def _check_or_write(rendered: list[RenderedFile], *, check: bool) -> int:
    expected_paths = {item.path.resolve() for item in rendered}
    existing_paths = set(COMMANDS_ROOT.glob("*.md"))
    stale_extra = sorted(p.resolve() for p in existing_paths if p.resolve() not in expected_paths)

    stale: list[Path] = []
    for item in rendered:
        current = item.path.read_text(encoding="utf-8") if item.path.exists() else None
        if current != item.content:
            stale.append(item.path.resolve())

    if check:
        if stale or stale_extra:
            print("Reference files are out of date. Run: python scripts/generate_reference.py")
            for path in sorted(stale):
                print(f"  MISSING/DIFF: {path.relative_to(REPO_ROOT)}")
            for path in stale_extra:
                print(f"  EXTRA: {path.relative_to(REPO_ROOT)}")
            return 1
        print("Reference files are up to date.")
        return 0

    for item in rendered:
        item.path.parent.mkdir(parents=True, exist_ok=True)
        item.path.write_text(item.content, encoding="utf-8")

    for path in stale_extra:
        path.unlink()

    print(f"Generated {len(rendered)} files under {REF_ROOT.relative_to(REPO_ROOT)}")
    return 0


def main() -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--check", action="store_true", help="Fail when generated files are stale.")
    args = parser.parse_args()
    rendered = _render()
    return _check_or_write(rendered, check=args.check)


if __name__ == "__main__":
    raise SystemExit(main())



import argparse
import configparser
from io import StringIO
import os
from pathlib import Path
import shutil
import sys
import tempfile


from pysisyphus import logger


CONFIG_DIR = Path(os.path.abspath(os.path.dirname(__file__)))
BASIS_LIB_DIR = Path(CONFIG_DIR / "basis_library")
LIB_DIR = Path(CONFIG_DIR / "geom_library")
WF_LIB_DIR = Path(CONFIG_DIR / "wf_library")
T_DEFAULT = 298.15  # Kelvin
p_DEFAULT = 101325  # Pascal
OUT_DIR_DEFAULT = str(Path(tempfile.gettempdir()) / "qm_calcs")
BEND_MIN_DEG = 15
LB_MIN_DEG = 175
DIHED_MAX_DEG = 175
L_MAX = 4
L_AUX_MAX = 5
AFIR_RMSD_THRESH = 0.25
DEFAULTS = {
    # .pysisyphusrc key: command
    "mwfn": "Multiwfn",
    "jmol": "jmol",
    "packmol": "packmol",
    "formchk": "formchk",
    "unfchk": "unfchk",
    "rwfdump": "rwfdump",
    # QC codes
    "orca": "orca",
    "orca5": "orca",
    "gaussian16": "g16",
    "wfoverlap": "wfoverlap.x",
    "openmolcas": "pymolcas",
    "gamess": "rungms",
    "xtb": "xtb",
    "mopac": "mopac",
}


# First try to read path to .pysisyphusrc from environment variable
try:
    pysisrc_env = os.getenv("PYSISRC", default=None)
    config_fn = Path(pysisrc_env).resolve()
    print(f"Read pysisyphus configuration from '{config_fn}'")
# Fallback to $HOME/.pysisyphusrc
except TypeError:
    config_fn = Path.home() / ".pysisyphusrc"

if not config_fn.is_file():
    print(f"Couldn't find configuration file. Expected it at '{config_fn}'.")

Config = configparser.ConfigParser()
read_fns = Config.read(config_fn)


def get_cmd(section, key="cmd", use_defaults=True):
    cmd = None
    msg = f" and no default was specified for '{section}'."

    # First we try to load the command from .pysisyphusrc
    try:
        cmd = Config[section][key]
        msg = "."
    # When the command is not available on .pysisyphusrc we check the defaults
    except KeyError:
        if use_defaults:
            default_key = section if key == "cmd" else key
            try:
                default_cmd = DEFAULTS[default_key]
                cmd = shutil.which(default_cmd)
                # 'msg' will only be printed when 'cmd' is None, so the msg is
                # always negative.
                msg = f" and default cmd='{default_cmd}' was not found in $PATH."
            except KeyError:
                pass

    if cmd is None:
        logger.warning(
            f"Failed to load '{key}' from [{section}] "
            f"in ~/.pysisyphusrc{msg}"
        )
        cmd = None
    return cmd


def detect_paths():
    config = configparser.ConfigParser()

    for k, v in DEFAULTS.items():
        print(f"{k: >16}: ... ", end="")
        if not (path := shutil.which(v)):
            print("not found.")
        else:
            path = Path(path).resolve()
            config[k] = {"cmd": str(path)}
            print(path)

    fp = StringIO()
    config.write(fp, space_around_delimiters=False)
    fp.seek(0)
    config_text = fp.read()
    return config_text


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--out", "-o", help="Filename to which the detected cmds are dumped."
    )

    return parser.parse_args(args)


def run_detect_paths():
    args = parse_args(sys.argv[1:])
    out = args.out

    config_text = detect_paths()
    if out:
        print("Dumped detected cmds to '{out}'")
        with open(out, "w") as handle:
            handle.write(config_text)
    else:
        print("\nExample .pysisyphusrc:\n")
        print(config_text)



import numpy as np

from pysisyphus.calculators.AnaPotBase import AnaPotBase


class AnaPot(AnaPotBase):
    def __init__(self, **kwargs):
        kwargs_ = {
            "V_str": "4 + 4.5*x - 4*y + x**2 + 2*y**2-2*x*y + x**4 - 2*x**2*y",
            "xlim": (-2, 2.5),
            "ylim": (0, 5),
            "levels": np.linspace(-3, 4, 80),
            "minima": ((-1.05274, 1.02776, 0), (1.94101, 3.85427, 0)),
            "saddles": ((0.6117313, 1.4929732, 0.0),),
        }
        kwargs_.update(kwargs)
        super().__init__(**kwargs_)

    def __str__(self):
        return "AnaPot calculator"



import importlib
import shutil

from pysisyphus.config import Config, DEFAULTS
from pysisyphus import logger

try:
    import pytest
except ModuleNotFoundError:
    logger.warning("pytest package could not be imported.")


"""Inspired by
https://github.com/MolSSI/QCEngine/blob/master/qcengine/testing.py

Maybe implement something like this to make test access easier
    https://stackoverflow.com/a/36601576/12486216
"""


_reason = "Calculator {} is not available."
_using_cache = dict()
# Python modules to be imported using importlib.import_module
IMPORT_DICT = {
    "dftd4": "dftd4",
    "pyscf": "pyscf",
    "dalton": "daltonproject",
    "qcengine": "qcengine",
    "openmm": "simtk.openmm",
    "thermoanalysis": "thermoanalysis",
    "pyxtb": "xtb.interface",
    "obabel": "openbabel.openbabel",
    "geodesic": "geodesic_interpolate",
}


def module_available(calculator):
    try:
        import_name = IMPORT_DICT[calculator]
    except KeyError:
        return False

    try:
        _ = importlib.import_module(import_name)
        available = True
    except (ModuleNotFoundError, ImportError):
        available = False
    return available


class DummyMark:
    def __init__(self, available):
        self.args = (available,)


def using(calculator, set_pytest_mark=True):
    """Calling disabling set_pytest_mark avoids a runtime dependency on pytest."""
    calculator = calculator.lower()

    if calculator not in _using_cache:
        available = False

        # Look into .pysisyphusrc first
        try:
            cmd = Config[calculator]["cmd"]
        except KeyError:
            # Try defaults last
            try:
                cmd = DEFAULTS[calculator]
            except KeyError:
                cmd = None

            # Look for dscf availability
            if calculator == "turbomole":
                cmd = "dscf"

        # Check if cmd is available on $PATH
        if cmd is not None:
            available = bool(shutil.which(cmd))
        # Handling native python packages from here
        else:
            available = module_available(calculator)

        reason = _reason.format(calculator)
        skipif = not available
        if set_pytest_mark:
            mark = pytest.mark.skipif(skipif, reason=reason)
        else:
            mark = DummyMark(skipif)
        _using_cache[calculator] = mark
    return _using_cache[calculator]


def available(calculator, **kwargs):
    # True when skipif is False
    return not using(calculator, **kwargs).args[0]



# [1] https://doi.org/10.1016/j.cplett.2004.07.079
#     The reaction pathway of a potential energy surface as curve
#     with induced tangent
#     Hirsch, Quapp, 2004


import numpy as np

from pysisyphus.calculators.AnaPotBase import AnaPotBase


class NFK(AnaPotBase):
    def __init__(self, C=0.06, **kwargs):
        """NFK surface of Hirsch and Quapp.

        See Fig. 4 in [1] and the appendix.
        """

        # Note that C is substituted in
        V_str = f"{C} * (x**2 + y**2)**2 + x*y - 9*exp(-(x-3)**2-y**2) - 9*exp(-(x+3)**2 - y**2)"
        kwargs_ = {
            "V_str": V_str,
            "xlim": (-3, 3),
            "ylim": (-3, 3),
            "levels": np.linspace(-7, 3.5, 70),
            "minima": ((2.71268103, -0.15093971, 0.0), (-2.7126810, 0.1509397, 0.0)),
            "saddles": ((0.0, 0.0, 0.0),),
        }
        kwargs_.update(kwargs)
        super().__init__(**kwargs_)

    def __str__(self):
        return "NFK calculator"


class ModNFK(NFK):
    def __init__(self, C=0.03, **kwargs):
        kwargs["minima"] = (
            (-2.84721930, 0.15702574, 0.0),
            (2.84721930, -0.15702574, 0.0),
        )
        super().__init__(C, **kwargs)

    def __str__(self):
        return "ModNFK calculator"



import argparse
import sys
import warnings

import numpy as np

try:
    from openbabel import pybel

    HAS_OPENBABEL = True
except ModuleNotFoundError:
    HAS_OPENBABEL = False


from pysisyphus.calculators import (
    Composite,
    HardSphere,
    PWHardSphere,
    TransTorque,
)
from pysisyphus.calculators.OBabel import OBabel
from pysisyphus.drivers.precon_pos_rot import (
    center_fragments,
    form_A,
    get_steps_to_active_atom_mean,
    get_which_frag,
    rotate_inplace,
    SteepestDescent,
)
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import align_coords, geom_loader
from pysisyphus.io import geom_to_crd_str
from pysisyphus.linalg import get_rot_mat_for_coords
from pysisyphus.optimizers.LBFGS import LBFGS
from pysisyphus.xyzloader import coords_to_trj


def merge_geoms(geom1, geom2, geom1_del=None, geom2_del=None, make_bonds=None):
    """Merge geom1 and geom2 while keeping the original coordinates.

    Supports deleting certain atoms.
    """

    if geom1_del is None:
        geom1_del = list()
    if geom2_del is None:
        geom2_del = list()

    geom1_del = np.array(geom1_del)
    geom2_del = np.array(geom2_del)
    # Update indices of atoms to be deleted in geom2 by the number of atoms in geom1,
    # as geom1 and geom2 will be merged.
    # If we want to delete atom 0 and atom 1 in geom2 and geom1 comprises 10 atoms,
    # then the updated indices in geom2 will be 0 + 10 = 10 and 1 + 10 = 11 in the
    # merged geometry.
    atom_num1 = len(geom1.atoms)
    geom2_del += atom_num1

    ndel1 = len(geom1_del)

    if make_bonds is not None:
        make_bonds = np.array(make_bonds, dtype=int)
        geom1_bond_inds, geom2_bond_inds = make_bonds.T
        geom2_bond_inds += atom_num1
        # Correct original bond indices given in make_bonds
        #
        # Determe how many atoms to be deleted lie below the atom(s) that are
        # used to form new bonds.
        corr1 = (geom1_del < geom1_bond_inds[:, None]).sum(axis=1)
        geom1_bond_inds -= corr1
        # Correct bond indices in geom2 by the number of deleted atoms in geom1.
        geom2_bond_inds -= ndel1
        corr2 = (geom2_del < geom2_bond_inds[:, None]).sum(axis=1)
        geom2_bond_inds -= corr2
        make_bonds_cor = np.stack((geom1_bond_inds, geom2_bond_inds), axis=1)
    else:
        make_bonds_cor = None

    union = geom1 + geom2
    union = union.del_atoms(list(geom1_del) + list(geom2_del))

    # Set appropriate fragments
    atom_num1 -= len(geom1_del)
    frag1 = np.arange(atom_num1)
    lig_atoms = geom2.atoms
    frag2 = np.arange(atom_num1, atom_num1 + len(lig_atoms) - len(geom2_del))
    union.fragments = {"geom1": frag1.tolist(), "geom2": frag2.tolist()}

    return union, make_bonds_cor


def hardsphere_merge(geom1, geom2):
    union = geom1 + geom2

    atom_num = len(geom1.atoms)
    # Set up lists containing the atom indices for the two fragments
    frag_lists = [
        [i for i, _ in enumerate(geom1.atoms)],
        [atom_num + i for i, _ in enumerate(geom2.atoms)],
    ]

    def get_hs(kappa=1.0):
        return HardSphere(
            union,
            frag_lists,
            permutations=True,
            kappa=kappa,
        )

    union.set_calculator(get_hs(1.0))
    opt_kwargs = {
        "max_cycles": 1000,
        "max_step": 0.5,
        "rms_force": 0.0005,
    }
    opt = SteepestDescent(union, **opt_kwargs)
    opt.run()
    return union


def prepare_merge(geom1, bond_diff, geom2=None, del1=None, del2=None, dump=False):
    bond_diff = np.array(bond_diff, dtype=int)
    if del1 is not None:
        geom1 = geom1.del_atoms(del1)

    if geom2:
        if del2 is not None:
            geom2 = geom2.del_atoms(del2)
        union = geom1 + geom2
        atom_num = len(geom1.atoms)
        # Set up lists containing the atom indices for the two fragments
        frag_lists = [
            [i for i, _ in enumerate(geom1.atoms)],
            [atom_num + i for i, _ in enumerate(geom2.atoms)],
        ]
        fragments = {"geom1": frag_lists[0], "geom2": frag_lists[1]}
        union.fragments = fragments
    else:
        union = geom1
    frag_lists = [union.fragments["geom1"], union.fragments["geom2"]]

    backup = list()

    def keep(comment):
        backup.append((union.cart_coords.copy(), comment))

    keep("Initial union")

    which_frag = get_which_frag(frag_lists)
    AR = form_A(frag_lists, which_frag, bond_diff)

    # Center fragments at their geometric average
    center_fragments(frag_lists, union)
    keep("Centered fragments")

    # Translate centroid of active atoms to origin
    alphas = get_steps_to_active_atom_mean(frag_lists, frag_lists, AR, union.coords3d)
    for frag, alpha in zip(frag_lists, alphas):
        union.coords3d[frag] += alpha
    keep("Shifted centroids to active atoms")

    def get_hs(kappa=1.0, **kwargs):
        return HardSphere(
            union,
            frag_lists,
            permutations=True,
            kappa=kappa,
            **kwargs,
        )

    union.set_calculator(get_hs(1.0))
    opt_kwargs = {
        "max_cycles": 1000,
        "max_step": 0.5,
        "rms_force": 0.0005,
    }
    opt = SteepestDescent(union, **opt_kwargs)
    opt.run()
    keep("Initial hardsphere optimization")

    # Corresponds to A.3 S_3 initial orientation of molecules in Habershon paper
    rotate_inplace(frag_lists, union, bond_diff)
    keep("Rotated after inital hardsphere optimization")

    sub_frags = [bond_diff[:, 0].tolist(), bond_diff[:, 1].tolist()]
    keys_calcs = {
        "hs": get_hs(100.0, radii_offset=3.0),
        "tt": TransTorque(frag_lists, frag_lists, AR, AR, kappa=2, do_trans=True),
        "pwhs": PWHardSphere(union, frag_lists, sub_frags=sub_frags, kappa=10.0),
    }
    comp = Composite("hs + tt + pwhs", keys_calcs, remove_translation=True)
    union.set_calculator(comp)

    max_cycles = 15_000
    max_dist = 50
    max_step = max_dist / max_cycles
    opt_kwargs = {
        "max_cycles": max_cycles,
        "max_step": max_step,
        "rms_force": 0.1,
        # "dump": True,
    }
    opt = SteepestDescent(union, **opt_kwargs)
    opt.run()
    keep("Hardsphere + TransTorque optimization")

    if dump:
        kept_coords, kept_comments = zip(*backup)
        align_coords(kept_coords)
        fn = "merge_dump_trj.xyz"
        coords_to_trj(fn, union.atoms, kept_coords, comments=kept_comments)

    return union


def merge_opt(union, bond_diff, ff="mmff94"):
    """Fragment merging along given bond by forcefield optimization."""

    geom1 = union.get_fragments("geom1")
    freeze = list(range(len(geom1.atoms)))

    # Create pybel.Molecule/OBMol to set the missing bonds
    mol = pybel.readstring("xyz", union.as_xyz())
    obmol = mol.OBMol
    for from_, to_ in bond_diff:
        obmol.AddBond(int(from_ + 1), int(to_ + 1), 1)

    # Use modified pybel.Molecule
    calc = OBabel(mol=mol, ff=ff)
    funion = union.copy()
    # Only releax second fragment
    funion.freeze_atoms = freeze
    funion.set_calculator(calc)

    opt = LBFGS(funion, max_cycles=1000, max_step=0.5, dump=False, print_every=25)
    opt.run()

    return funion


def align_on_subset(geom1, union, del1=None):
    """Align 'union' onto subset of 'geom1'"""

    # Delete some coordinates (atoms) in geom1, that are not present in union
    coords3d_1 = geom1.coords3d.copy()
    atoms1 = geom1.atoms
    if del1 is not None:
        atoms1 = [atom for i, atom in enumerate(atoms1) if i not in del1]
        coords3d_1 = np.delete(coords3d_1, del1, axis=0)
    atoms1 = tuple(atoms1)
    num1 = coords3d_1.shape[0]
    # Restrict length of union to length of coords3d_1
    coords3d_2 = union.coords3d.copy()
    coords3d_2_subset = coords3d_2[:num1]
    atoms2 = union.atoms
    assert atoms1 == tuple(atoms2[:num1])

    *_, rot_mat = get_rot_mat_for_coords(coords3d_1, coords3d_2_subset)

    # Align merged system
    coords3d_2_aligned = (coords3d_2 - coords3d_2.mean(axis=0)[None, :]).dot(rot_mat)

    # Translate aligned system so that centroids of subsets match
    centroid_1 = coords3d_1.mean(axis=0)
    centroid_2 = coords3d_2_aligned[:num1].mean(axis=0)
    coords3d_2_aligned += centroid_1 - centroid_2

    aligned = Geometry(atoms2, coords3d_2_aligned)
    subset = Geometry(atoms2[:num1], coords3d_2_aligned[:num1])
    rest = Geometry(atoms2[num1:], coords3d_2_aligned[num1:])
    return aligned, subset, rest


def merge_with_frozen_geom(
    frozen_geom, lig_geom, make_bonds, frozen_del, lig_del, ff="mmff94"
):
    union, make_bonds_cor = merge_geoms(
        frozen_geom, lig_geom, frozen_del, lig_del, make_bonds
    )
    atoms = union.atoms
    print("Docking to form bonds:")
    for i, (from_, to_) in enumerate(make_bonds_cor):
        from_atom = atoms[from_]
        to_atom = atoms[to_]
        print(f"\t{i:02d} {from_atom}{from_}-{to_atom}{to_}")

    union = prepare_merge(union, make_bonds_cor, dump=True)
    if HAS_OPENBABEL:
        opt_union = merge_opt(union, make_bonds_cor, ff=ff)
    else:
        warnings.warn(f"Could not import openbabel. Skipping '{ff}' optimization.")
        opt_union = union
    # aligned: whole system
    # subset: frozen_geom - deleted atoms
    # rest: aligned ligand
    aligned, subset, rest = align_on_subset(frozen_geom, opt_union, frozen_del)
    return aligned, subset, rest


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument("frozen_fn", help="Filename of the frozen geometry.")
    parser.add_argument("lig_fn", help="Filename of the ligand to be merged.")
    parser.add_argument(
        "--frozen-del",
        nargs="+",
        type=int,
        help="0-based atom indices to be deleted from frozen_fn.",
    )
    parser.add_argument(
        "--lig-del",
        nargs="+",
        type=int,
        help="0-based atom indices to be deleted from lig_fn.",
    )
    parser.add_argument(
        "--make-bonds",
        nargs="+",
        type=int,
        help="0-based indices of atom pairs (frozen, lig), forming bonds "
        "in the merged geometry. lig indices should start at 0.",
    )
    parser.add_argument("--res", default="LIG")
    parser.add_argument("--resno", type=int)

    return parser.parse_args(args)


def run_merge():
    args = parse_args(sys.argv[1:])

    frozen_fn = args.frozen_fn
    lig_fn = args.lig_fn
    protein_pdb = frozen_fn
    lig_pdb = lig_fn

    frozen_geom = geom_loader(protein_pdb)
    lig_geom = geom_loader(lig_pdb)

    prot_del = args.frozen_del
    lig_del = args.lig_del
    make_bonds = args.make_bonds
    make_bonds = np.array(make_bonds, dtype=int).reshape(-1, 2)
    aligned, subset, rest = merge_with_frozen_geom(
        frozen_geom, lig_geom, make_bonds, prot_del, lig_del
    )

    trj_fn = "merged_trj.xyz"
    trj = "\n".join([geom.as_xyz() for geom in (aligned, rest)])
    with open(trj_fn, "w") as handle:
        handle.write(trj)
    print(f"Dumped geometries to '{trj_fn}'.")

    res = args.res
    resno = args.resno
    if (res is not None) and (resno is not None):
        crd_str = geom_to_crd_str(
            rest, res=res, resno=resno, ref_atoms=lig_geom.atoms, del_atoms=lig_del
        )
        crd_fn = "lig_aligned.crd"
        with open(crd_fn, "w") as handle:
            handle.write(crd_str)
        print(f"Dumped ligand coordinates to  to '{crd_fn}'.")


if __name__ == "__main__":
    run_merge()



import numpy as np

from pysisyphus.constants import BOHR2ANG
from pysisyphus.Geometry import Geometry


def parse_coord_line(line):
    x, y, z, atom, *_ = line.strip().split()
    return (x, y, z), atom


def parse_sdf(text):
    lines = text.split("\n")
    # title, program, comment = lines[:3]
    count = lines[3]
    atoms, *_ = count.split()
    atoms = int(atoms)

    coord_lines = lines[4:4+int(atoms)]
    coords, atoms = zip(*[parse_coord_line(cl) for cl in coord_lines])
    coords = np.array(coords, dtype=float)
    coords /= BOHR2ANG
    return atoms, coords


def geom_from_sdf(text, **kwargs):
    atoms, coords = parse_sdf(text)
    geom = Geometry(atoms, coords, **kwargs)
    return geom



import matplotlib.pyplot as plt
import numpy as np


class ParamPlot:

    def __init__(self, coords_list, param1_inds, param2_inds):
        self.coords_list = coords_list
        self.coords_list_3d = [c.reshape(-1, 3) for c in coords_list]
        self.p1inds = param1_inds
        self.p2inds = param2_inds

        self.p1 = [self.get_param(c, param1_inds) for c in self.coords_list_3d]
        self.p2 = [self.get_param(c, param2_inds) for c in self.coords_list_3d]

        self.fig, self.ax = plt.subplots()

    def calc_bond(self, coords, ind1, ind2):
        return np.linalg.norm(coords[ind1] - coords[ind2])

    def calc_angle(self, coords, ind1, ind2, ind3):
        vec1 = coords[ind1] - coords[ind2]
        vec2 = coords[ind3] - coords[ind2]
        vec1n = np.linalg.norm(vec1)
        vec2n = np.linalg.norm(vec2)
        dotp = np.dot(vec1, vec2)
        radians = np.arccos(dotp / (vec1n * vec2n))
        return radians * 180 / np.pi

    def get_param(self, coords, param_inds):
        if len(param_inds) == 2:
            return self.calc_bond(coords, *param_inds)
        elif len(param_inds) == 3:
            return self.calc_angle(coords, *param_inds)
        else:
            raise Exception

    def plot(self):
        self.ax.plot(self.p1, self.p2, "ro", ls="--")
        self.ax.set_xlabel(str(self.p1inds))
        self.ax.set_ylabel(str(self.p2inds))

    def show(self):
        plt.tight_layout()
        plt.show()

    def save(self, path, prefix):
        out_fn = str(path / (prefix + ".pdf"))
        self.fig.tight_layout()
        self.fig.savefig(out_fn)

    def save_coords(self, path, prefix):
        np.savetxt(path / (prefix + ".coords"), self.coords_list)



from math import prod

import numpy as np
from numpy.typing import NDArray
from scipy.special import factorial2

from pysisyphus.wavefunction.helpers import canonical_order


# @functools.cache
def get_lmn_factors(L: int):
    lmns = canonical_order(L)
    lmn_factors = np.zeros(len(lmns))
    for i, lmn in enumerate(lmns):
        lmn_factors[i] = prod([factorial2(2 * am - 1) for am in lmn])
    lmn_factors = 1 / np.sqrt(lmn_factors)
    return lmn_factors


def norm_cgto_lmn(coeffs: NDArray[float], exps: NDArray[float], L: int):
    N = 0.0
    for i, expi in enumerate(exps):
        for j, expj in enumerate(exps):
            tmp = coeffs[i] * coeffs[j] / (expi + expj) ** (L + 1.5)
            tmp *= np.sqrt(expi * expj) ** (L + 1.5)
            N += tmp
    N = np.sqrt(exps ** (L + 1.5) / (np.pi**1.5 / 2**L * N))
    # Or w/ array broadccasting
    # N = coeffs[None, :] * coeffs[:, None] / (exps[None, :] + exps[:, None]) ** (L + 1.5)
    # N = (N * np.sqrt((exps[None, :] * exps[:, None]) ** (L + 1.5))).sum()
    # N = np.sqrt(exps ** (L + 1.5) / (np.pi**1.5 / 2**L * N))

    mod_coeffs = N * coeffs
    lmn_factors = get_lmn_factors(L)

    return mod_coeffs, lmn_factors



import numpy as np

from pysisyphus.helpers import do_final_hessian
from pysisyphus.helpers_pure import eigval_to_wavenumber
from pysisyphus.irc.IRC import IRC
from pysisyphus.optimizers.hessian_updates import bofill_update


class ModeKill(IRC):

    def __init__(self, geometry, kill_inds, nu_thresh=-5.,
                 do_hess=True, hessian_update=True, **kwargs):
        # Use tight convergence criteria so the IRC/ModeKill doesn't
        # ent prematurely.
        super().__init__(geometry, downhill=True, rms_grad_thresh=1e-6,
                         hessian_init="calc", **kwargs)

        assert self.geometry.coord_type == "cart"

        self.kill_inds = np.array(kill_inds, dtype=int)
        self.nu_thresh = float(nu_thresh)
        self.do_hess = do_hess
        self.hessian_update = hessian_update

        self.fmt = {"float": lambda f: f"{f:.4f}",}
        self.ovlp_thresh = .3
        self.indices = np.arange(self.geometry.coords.size)
        self.neg_nus = list()

    def prepare(self, *args, **kwargs):
        super().prepare(*args, **kwargs)

        self.update_mw_down_step()

    def update_mw_down_step(self):
        w, v = np.linalg.eigh(self.mw_hessian)

        self.kill_modes = v[:,self.kill_inds]
        nus = eigval_to_wavenumber(w)
        assert all(nus[self.kill_inds] < self.nu_thresh), \
              "ModeKill is intended for removal of imaginary frequencies " \
             f"below {self.nu_thresh} cm⁻¹! The specified indices " \
             f"{self.kill_inds} contain modes with positive frequencies " \
             f"({nus[self.kill_inds]} cm⁻¹). Please choose different kill_inds!"

        """After diagonalization of the mass-weighted hessian the signs
        of the eigenvectors are arbitrary.
        We determine the correct sign of the eigenvector(s) from its
        overlap(s) with the gradient.
        To decrease the overall energy we have to step against the
        gradient, so the overlap of gradient and the respective eigen-
        vector(s) must be negative.
        If an overlap is positive we flip the sign of the corresponding
        eigenvector.
        """
        mw_grad = self.mw_gradient
        mw_grad_normed = mw_grad / np.linalg.norm(mw_grad)
        overlaps = np.einsum("ij,i->j", self.kill_modes, mw_grad_normed)
        self.log("Overlaps between gradient and eigenvectors:")
        self.log(overlaps)
        flip = overlaps > 0
        self.log("Eigenvector signs to be flipped:")
        self.log(str(flip))
        self.kill_modes[:,flip] *= -1
        # Create the step as the sum of the downhill steps along the modes
        # to remove.
        self.mw_down_step = (self.step_length*self.kill_modes).sum(axis=1)

    def step(self):
        if self.hessian_update and (self.cur_cycle > 0):
            # Hessian update with mass-weighted values
            dx = self.irc_mw_coords[-1] - self.irc_mw_coords[-2]
            dg = (self.irc_mw_gradients[-1] - self.irc_mw_gradients[-2])
            d_mw_H, key = bofill_update(self.mw_hessian, dx, dg)
            self.mw_hessian += d_mw_H

            # norm(dx) is probably self.step_length ;)
            norm_dx = np.linalg.norm(dx)
            norm_dg = np.linalg.norm(dg)
            self.log(f"Did {key} hessian update: norm(dx)={norm_dx:.4e}, "
                     f"norm(dg)={norm_dg:.4e}."
            )
        else:
            # Recalculate exact hessian
            self.mw_hessian = self.geometry.mw_hessian
            self.log("Recalculated exact hessian.")

        w, v = np.linalg.eigh(self.mw_hessian)
        # Overlaps between current normal modes and the modes we want to
        # remove.
        overlaps = np.abs(np.einsum("ij,ik->jk", self.kill_modes, v))
        self.log(f"Overlaps between original modes and current modes:")
        # overlaps contains one row per mode to remove
        for i, ovlps in enumerate(overlaps):
            above_thresh = ovlps > self.ovlp_thresh
            ovlp_str = " ".join(
                [f"{i:02d}: {o:.4f}" for i, o in zip(self.indices[above_thresh],
                                                     ovlps[above_thresh])
                ]
            )
            self.log(f"Org. mode {i:02d}:\n\t\t\t{ovlp_str}")

        nus = eigval_to_wavenumber(w)
        neg_inds = nus <= self.nu_thresh
        neg_nus = nus[neg_inds]
        self.neg_nus.append(neg_nus)
        self.log(f"Wavenumbers of imaginary modes (<= {self.nu_thresh} cm⁻¹):")
        self.log(f"{neg_nus} cm⁻¹")

        # Check if any eigenvalues became positive. If so remove them and update
        # the step. If no mode to remove is left we are finished and can signal
        # convergence.
        argmax = overlaps.argmax(axis=1)
        eigvals = w[argmax]
        pos_eigvals = eigvals > 0
        if any(pos_eigvals):
            # Only keep negative eigenvalues.
            flipped = self.kill_inds[pos_eigvals]
            self.kill_inds = self.kill_inds[~pos_eigvals]
            self.update_mw_down_step()
            self.log(f"Eigenvalue(s) of mode(s) {flipped} became positive!")
        self.converged = len(self.kill_inds) == 0

        if not self.converged:
            self.mw_coords += self.mw_down_step

    def postprocess(self):
        super().postprocess()

        if self.do_hess:
            print()
            do_final_hessian(self.geometry)

    def get_additional_print(self):
        neg_nus = np.array2string(self.neg_nus[-1], precision=2)
        return f"\timag. ῦ: {neg_nus} cm⁻¹"



from dataclasses import dataclass
import itertools as it
import pickle
from typing import Dict, List

import numpy as np
from numpy.typing import NDArray

from pysisyphus.Geometry import Geometry
from pysisyphus.io.pdb import parse_pdb
from pysisyphus.io.psf import parse_psf


@dataclass
class Atom:
    id: int
    segment: str
    resid: int
    resname: str
    name: str
    type: str
    charge: float
    mass: float
    coords: NDArray[float]
    element: str

    @staticmethod
    def from_psf_line(line, coords, element):
        return Atom(
            id=line["id"],
            segment=line["segment"],
            resid=line["resid"],
            resname=line["resname"],
            name=line["atom_name"],
            type=line["atom_type"],
            charge=line["charge"],
            mass=line["mass"],
            coords=coords,
            element=element,
        )


@dataclass
class Residue:
    id: int
    name: str
    segment: str
    atoms: List[Atom]

    @staticmethod
    def from_psf_lines(lines, coords3d, elements):
        assert len(lines) == len(coords3d) == len(elements)
        atoms = [
            Atom.from_psf_line(line, coords, element)
            for line, coords, element in zip(lines, coords3d, elements)
        ]
        atom0 = atoms[0]
        resid0 = atom0.resid
        resname0 = atom0.resname
        segment0 = atom0.segment
        if len(lines) > 1:
            resids, resnames = zip(*[(atom.resid, atom.resname) for atom in atoms[1:]])
            assert all([resname == resname0 for resname in resnames])
            assert all([resid == resid0 for resid in resids])
        return Residue(resid0, resname0, segment0, atoms)

    @property
    def key(self):
        return self.segment, self.id

    @property
    def charge(self):
        charge = sum([atom.charge for atom in self.atoms])
        # assert abs(charge % 1) <= 1e-10  # I guess charges must not be integer ...
        return charge

    @property
    def masses(self):
        return np.array([atom.mass for atom in self.atoms])

    @property
    def total_mass(self):
        return sum(self.masses)

    @property
    def atom_indices(self):
        return [atom.id for atom in self.atoms]

    @property
    def elements_coords3d(self):
        elements = [atom.element for atom in self.atoms]
        return elements, self.coords3d

    @property
    def coords3d(self):
        return np.array([atom.coords for atom in self.atoms])

    @property
    def center_of_mass(self):
        return (
            1 / self.total_mass * np.sum(self.coords3d * self.masses[:, None], axis=0)
        )

    @property
    def com(self):
        return self.center_of_mass

    def __len__(self):
        return len(self.atoms)

    def __str__(self):
        return f"{self.resname}{self.resid}"


"""
@dataclass
class Residues:
    residues: Dict
    psf_data: Dict

    def as_geom(self, with_link_atoms=True):
        geom = geom_from_residues(self.residues)
        atoms = geom.atoms
        coords3d = geom.coords3d
        bonds = np.array(self.psf_data["nbond"]["inds"], dtype=int).reshape(-1, 2)
        if with_link_atoms:
            link_hosts, link_atoms, link_coords3d = link_atoms_for_residues(
                self.residues, bonds, coords3d, atom_map
            )
        atoms += link_atoms
        coords3d = np.concatenate((coords3d, link_coords3d), axis=0)
        geom = Geometry(atoms, coords3d)
        return geom
"""


def residues_from_psf(psf_data, atoms, coords3d, atom_map):
    psf_atoms = psf_data["atoms"]
    res_key = lambda atom: (atom["segment"], atom["resid"])
    psf_atoms = sorted(psf_atoms, key=res_key)
    psf_atoms_by_res = it.groupby(psf_atoms, key=res_key)
    residues = dict()
    for _, g in psf_atoms_by_res:
        res_atoms = sorted(g, key=lambda atom: atom["id"])
        atom_inds = [atom_map[atom["id"]] for atom in res_atoms]
        res_coords3d = coords3d[atom_inds]
        res_elements = [atoms[i] for i in atom_inds]
        res = Residue.from_psf_lines(res_atoms, res_coords3d, res_elements)
        residues[res.key] = res
    return residues


def residues_within_dist(
    residues,
    within_resid,
    within_dist,
    kind="com",
):
    def com_within():
        coms = {key: res.com for key, res in residues.items()}
        ref_com = coms[within_resid]
        res_ids_below_dist = [
            key
            for key, res in residues.items()
            if np.linalg.norm(coms[res.key] - ref_com) <= within_dist
        ]
        return res_ids_below_dist

    def atom_within():
        ref_res = residues[within_resid]
        ref_coords3d = ref_res.coords3d
        res_ids_below_dist = list()
        for key, res in residues.items():
            dist_vecs = ref_coords3d[:, None, :] - res.coords3d
            dists = np.linalg.norm(dist_vecs, axis=2)
            if (dists <= within_dist).any():
                res_ids_below_dist.append(key)
        return res_ids_below_dist

    if kind == "com":
        within_resid = com_within()
    elif kind == "atom":
        within_resid = atom_within()
        pass
    else:
        raise Exception(f"{kind=} is not supported!")

    residues_within = [residues[resid] for resid in within_resid]
    return residues_within


def geom_from_residues(residues):
    atoms = list()
    coords3d = np.zeros((sum([len(res) for res in residues]), 3))
    i = 0
    for res in residues:
        res_elements, res_coords3d = res.elements_coords3d
        len_res = len(res)
        coords3d[i : i + len_res] = res_coords3d
        atoms.extend(res_elements)
        i += len_res
    return Geometry(atoms, coords3d)


def link_atoms_for_residues(
    residues, bonds, coords3d, atom_map, link_element="H", g=0.709
):
    atom_inds = list(it.chain(*[res.atom_indices for res in residues]))

    bond_dict = dict()
    for bond in bonds:
        from_, to_ = bond
        bond_dict.setdefault(from_, set()).add(to_)
        bond_dict.setdefault(to_, set()).add(from_)
    atom_set = set(atom_inds)

    cut_bonds = list()
    # Check all bonds from residue-atoms. Determine which bonds are cut.
    for atom in atom_inds:
        try:
            cut_bonds_with = bond_dict[atom] - atom_set
        # Single ions/atoms may not have any bonds.
        except KeyError:
            cut_bonds_with = list()
        for cbw in cut_bonds_with:
            cut_bonds.append((atom, cbw))

    link_atoms = list()
    link_hosts = np.zeros(len(cut_bonds))
    link_coords3d = np.zeros((len(cut_bonds), 3))
    for i, cut_bond in enumerate(cut_bonds):
        from_, to_ = cut_bond
        link_atoms.append(link_element)
        link_hosts[i] = to_
        from_coords = coords3d[atom_map[from_]]
        to_coords = coords3d[atom_map[to_]]
        bond_vec = to_coords - from_coords
        link_coords3d[i] = from_coords + g * bond_vec
    link_atoms = tuple(link_atoms)
    return link_hosts, link_atoms, link_coords3d


def load_psf(psf_fn):
    if str(psf_fn).lower().endswith(".psf"):
        psf_data = parse_psf(psf_fn)
    else:
        with open(psf_fn, "rb") as handle:
            psf_data = pickle.load(handle)
    return psf_data


def cluster_from_psf_pdb(
    # psf_fn, pdb_fn, within_resid=None, within_dist=0.0, ref_residues=None, kind="atom,"
    psf_data,
    pdb_fn,
    within_resid=None,
    within_dist=0.0,
    ref_residues=None,
    kind="atom",
):
    atoms, coords, _, atom_map = parse_pdb(pdb_fn)
    coords3d = coords.reshape(-1, 3)
    print("Loaded PDB data.")

    residues = residues_from_psf(psf_data, atoms, coords3d, atom_map)

    # Select according to COM distance
    if within_resid and within_dist:
        # sel_residues = residues_within_com_dist(
        sel_residues = residues_within_dist(
            residues,
            within_resid,
            within_dist,
            kind=kind,
        )
    # Select residues according to provided ref_residues.
    elif ref_residues:
        sel_residues = [residues[ref_res.key] for ref_res in ref_residues]
    # Or complain!
    else:
        raise Exception(
            "Either 'within_resid' and 'within_dist' or 'residues' must be given!"
        )

    geom = geom_from_residues(sel_residues)
    print("Create cluster geometry.")
    bonds = np.array(psf_data["nbond"]["inds"], dtype=int).reshape(-1, 2)
    link_hosts, link_atoms, link_coords3d = link_atoms_for_residues(
        sel_residues, bonds, coords3d, atom_map
    )
    sat_geom = Geometry(
        geom.atoms + link_atoms, np.concatenate((geom.coords3d, link_coords3d), axis=0)
    )
    print("Created satured geometry with link atoms.")

    # Determine backbone atoms and/or link atoms and report their indices, so they can
    # be restrained in subsequent RMSD-biased optimizations. Also report the distance
    # between their positions and the COM of the reference residue.
    backbone_names = {"CA", "C", "O", "N"}  # HN, HA not included
    i = 0
    backbone_inds = list()
    backbone_com_dists = list()
    ref_com = residues[within_resid].com
    for res in sel_residues:
        for atom in res.atoms:
            if atom.name in backbone_names:
                backbone_inds.append(i)
                bb_dist = np.linalg.norm(ref_com - atom.coords)
                backbone_com_dists.append(bb_dist)
                print(
                    f"{res.name: >4s}{res.id: <5d}, {atom.element: >2s}{atom.id: <5d}, "
                    f"type={atom.name: >4s}, id={i: >5d}, {bb_dist: >8.4f} au"
                )
            i += 1
    backbone_inds = np.array(backbone_inds, dtype=int)
    backbone_com_dists = np.array(backbone_com_dists, dtype=float)

    return sat_geom, sel_residues, backbone_inds, backbone_com_dists



from copy import copy
import logging
import sys

from distributed import Client
import numpy as np
from scipy.interpolate import interp1d, splprep, splev

from pysisyphus.helpers import align_coords, get_coords_diffs
from pysisyphus.helpers_pure import hash_arr
from pysisyphus.modefollow import geom_lanczos


# [1] http://dx.doi.org/10.1063/1.1323224


class ChainOfStates:
    logger = logging.getLogger("cos")
    valid_coord_types = ("cart", "cartesian", "dlc")

    def __init__(
        self,
        images,
        fix_first=True,
        fix_last=True,
        align_fixed=True,
        climb=False,
        climb_rms=5e-3,
        climb_lanczos=False,
        climb_lanczos_rms=5e-3,
        climb_fixed=True,
        energy_min_mix=False,
        scheduler=None,
        progress=False,
    ):
        assert len(images) >= 2, "Need at least 2 images!"
        self.images = list(images)
        self.fix_first = fix_first
        self.fix_last = fix_last
        self.align_fixed = align_fixed
        self.climb = climb
        self.climb_rms = climb_rms
        self.climb_lanczos = climb_lanczos
        self.climb_fixed = climb_fixed
        self.energy_min_mix = energy_min_mix
        # Must not be lower than climb_rms
        self.climb_lanczos_rms = min(self.climb_rms, climb_lanczos_rms)
        self.scheduler = scheduler
        self.progress = progress

        self._coords = None
        self._forces = None
        self._energy = None
        self.counter = 0
        self.coords_length = self.images[0].coords.size
        self.cart_coords_length = self.images[0].cart_coords.size
        self.zero_vec = np.zeros(self.coords_length)

        self.coords_list = list()
        self.forces_list = list()
        self.all_energies = list()
        self.all_true_forces = list()
        self.lanczos_tangents = dict()
        self.prev_lanczos_hash = None

        # Start climbing immediateley with climb_rms == -1
        self.started_climbing = self.climb_rms == -1
        if self.started_climbing:
            self.log("Will start climbing immediately.")
        self.started_climbing_lanczos = False
        self.fixed_climb_indices = None
        # Use original forces for these images
        self.org_forces_indices = list()

        img0 = self.images[0]
        self.image_atoms = copy(img0.atoms)
        self.coord_type = img0.coord_type
        assert (
            self.coord_type in self.valid_coord_types
        ), f"Invalid coord_type! Supported types are: {self.valid_coord_types}"
        assert all(
            [img.coord_type == self.coord_type for img in self.images]
        ), "coord_type of images differ!"
        try:
            self.typed_prims = img0.internal.typed_prims
        except AttributeError:
            self.typed_prims = None

    @property
    def calculator(self):
        try:
            calc = self.images[0].calculator
        except IndexError:
            calc = None
        return calc

    def log(self, message):
        self.logger.debug(f"Counter {self.counter+1:03d}, {message}")

    def get_fixed_indices(self):
        fixed = list()
        if self.fix_first:
            fixed.append(0)
        if self.fix_last:
            fixed.append(len(self.images) - 1)
        return fixed

    @property
    def moving_indices(self):
        """Returns the indices of the images that aren't fixed and can be
        optimized."""
        fixed = self.get_fixed_indices()
        return [i for i in range(len(self.images)) if i not in fixed]

    @property
    def last_index(self):
        return len(self.images) - 1

    @property
    def moving_images(self):
        return [self.images[i] for i in self.moving_indices]

    @property
    def max_image_num(self):
        return len(self.images)

    @property
    def image_inds(self):
        return list(range(self.max_image_num))

    def zero_fixed_vector(self, vector):
        fixed = self.get_fixed_indices()
        for i in fixed:
            vector[i] = self.zero_vec
        return vector

    def clear(self):
        self._energy = None
        self._forces = None
        self._hessian = None
        try:
            self._tangents = None
        except AttributeError:
            # TODO: move this to another logging level?!
            self.log("There are no tangents to reset.")

    # @property
    # def freeze_atoms(self):
    # image_freeze_atoms = [image.freeze_atoms for image in self.images]
    # lens = [len(fa) for fa in image_freeze_atoms]
    # len0 = lens[0]
    # assert all([len_ == len0 for len_ in lens])
    # return image_freeze_atoms[0]

    @property
    def atoms(self):
        atoms_ = self.images[0].atoms
        return len(self.images) * atoms_

    def set_vector(self, name, vector, clear=False):
        vec_per_image = vector.reshape(-1, self.coords_length)
        assert len(self.images) == len(vec_per_image)
        for i in self.moving_indices:
            setattr(self.images[i], name, vec_per_image[i])
        if clear:
            self.clear()

    @property
    def coords(self):
        """Return a flat 1d array containing the coordinates of all images."""
        all_coords = [image.coords for image in self.images]
        # Note: why does this getter set self._coords? ... I wrote this line 6 years ago.
        self._coords = np.concatenate(all_coords)
        return self._coords

    @coords.setter
    def coords(self, coords):
        """Distribute the flat 1d coords array over all images."""
        self.set_vector("coords", coords, clear=True)

    @property
    def cart_coords(self):
        """Return a flat 1d array containing the cartesian coordinates of all
        images."""
        return np.concatenate([image.cart_coords for image in self.images])

    @property
    def coords3d(self):
        assert self.images[0].coord_type == "cart"
        return self.coords.reshape(-1, 3)

    @property
    def image_coords(self):
        return np.array([image.coords for image in self.images])

    def set_coords_at(self, i, coords):
        """Called from helpers.procrustes with cartesian coordinates.
        Then tries to set cartesian coordinate as self.images[i].coords
        which will raise an error when coord_type != "cart".
        """
        assert self.images[i].coord_type in ("cart", "cartesian"), (
            "ChainOfStates.set_coords_at() has to be reworked to support "
            "internal coordiantes. Try to set 'align: False' in the 'opt' "
            "section of the .yaml input file."
        )
        if i in self.moving_indices:
            self.images[i].coords = coords
        # When dealing with a fixed image don't set coords through the
        # property, which would result in resetting the image's calculated
        # data. Instead, assign coords directly. This only occurs when
        # aligning the fixed images.
        elif self.align_fixed:
            self.images[i]._coords = coords

    @property
    def energy(self):
        self._energy = np.array([image.energy for image in self.images])
        return self._energy

    @energy.setter
    def energy(self, energies):
        """This is needed for some optimizers like CG and BFGS."""
        assert len(self.images) == len(energies)
        for i in self.moving_indices:
            self.images[i].energy = energies[i]

        self._energy = energies

    def par_image_calc(self, image):
        image.calc_energy_and_forces()
        return image

    def set_images(self, indices, images):
        for ind, image in zip(indices, images):
            self.images[ind] = image

    def concurrent_force_calcs(self, images_to_calculate, image_indices):
        client = self.get_dask_client()
        self.log(client)

        # save original pals to restore them later
        orig_pal = images_to_calculate[0].calculator.pal

        # divide pal of each image by the number of workers available or available images
        # number of workers available
        n_workers = len(client.scheduler_info()["workers"])
        # number of images to calculate
        n_images = len(images_to_calculate)
        # divide pal by the number of workers (or 1 if more workers)
        new_pal = max(1, orig_pal // n_workers)

        # split images to calculate into batches of n_workers and set pal of each image
        n_batches = n_images // n_workers
        for i in range(0, n_batches):
            for j in range(i * n_workers, (i + 1) * n_workers):
                images_to_calculate[j].calculator.pal = new_pal

        # distribute the pals among the remaining images
        n_last_batch = n_images % n_workers
        if n_last_batch > 0:
            # divide pal by the remainder
            new_pal = max(1, orig_pal // n_last_batch)
            for i in range(n_batches * n_workers, n_images):
                images_to_calculate[i].calculator.pal = new_pal

        # map images to workers
        image_futures = client.map(self.par_image_calc, images_to_calculate)
        # set images to the results of the calculations
        self.set_images(image_indices, client.gather(image_futures))

        # Restore original pals
        for i in range(0, n_images):
            self.images[i].calculator.pal = orig_pal

    def calculate_forces(self):
        # Determine the number of images for which we have to do calculations.
        # There may also be calculations for fixed images, as they need an
        # energy value. But every fixed image only needs a calculation once.
        images_to_calculate = self.moving_images
        image_indices = self.moving_indices
        if self.fix_first and (self.images[0]._energy is None):
            images_to_calculate = [self.images[0]] + images_to_calculate
            image_indices = [0] + list(image_indices)
        if self.fix_last and (self.images[-1]._energy is None):
            images_to_calculate = images_to_calculate + [self.images[-1]]
            image_indices = list(image_indices) + [-1]
        assert len(images_to_calculate) <= len(self.images)

        # Parallel calculation with dask
        if self.scheduler:
            self.concurrent_force_calcs(images_to_calculate, image_indices)
        # Serial calculation
        else:
            for image in images_to_calculate:
                image.calc_energy_and_forces()
                # Poor mans progress bar ;)
                if self.progress:
                    print(".", end="")
                    sys.stdout.flush()
            if self.progress:
                print("\r", end="")
        self.set_zero_forces_for_fixed_images()
        self.counter += 1

        if self.energy_min_mix:
            # Will be None for calculators that already mix
            all_energies = np.array([image.all_energies for image in self.images])
            energy_diffs = np.diff(all_energies, axis=1).flatten()
            calc_inds = all_energies.argmin(axis=1)
            mix_at = []
            for i, calc_ind in enumerate(calc_inds[:-1]):
                next_ind = calc_inds[i + 1]
                if (
                    (calc_ind != next_ind)
                    and (i not in self.org_forces_indices)
                    and (i + 1 not in self.org_forces_indices)
                ):
                    min_diff_offset = energy_diffs[[i, i + 1]].argmin()
                    mix_at.append(i + min_diff_offset)

            for ind in mix_at:
                self.images[ind].calculator.mix = True
                # Recalculate correct energy and forces
                print(
                    f"Switch after calc_ind={calc_ind} at index {ind}. Recalculating."
                )
                self.images[ind].calc_energy_and_forces()
                self.org_forces_indices.append(ind)
                calc_ind = calc_inds[ind]

        energies = [image.energy for image in self.images]
        forces = np.array([image.forces for image in self.images])
        self.all_energies.append(energies)
        self.all_true_forces.append(forces)

        return {
            "energies": energies,
            "forces": forces,
        }

    @property
    def forces(self):
        self.set_zero_forces_for_fixed_images()
        forces = [image.forces for image in self.images]
        self._forces = np.concatenate(forces)
        self.counter += 1
        return self._forces

    @forces.setter
    def forces(self, forces):
        self.set_vector("forces", forces)

    @property
    def perpendicular_forces(self):
        indices = range(len(self.images))
        perp_forces = [self.get_perpendicular_forces(i) for i in indices]
        return np.array(perp_forces).flatten()

    def get_perpendicular_forces(self, i):
        """[1] Eq. 12"""
        # Our goal in optimizing a ChainOfStates is minimizing the
        # perpendicular force. Always return zero perpendicular
        # forces for fixed images, so that they don't interfere
        # with the convergence check.
        if i not in self.moving_indices:
            return self.zero_vec

        forces = self.images[i].forces
        tangent = self.get_tangent(i)
        perp_forces = forces - forces.dot(tangent) * tangent
        return perp_forces

    @property
    def gradient(self):
        return -self.forces

    @gradient.setter
    def gradient(self, gradient):
        self.forces = -gradient

    @property
    def masses_rep(self):
        return np.array([image.masses_rep for image in self.images]).flatten()

    @property
    def results(self):
        tmp_results = list()
        for image in self.images:
            res = image.results
            res["coords"] = image.coords
            res["cart_coords"] = image.cart_coords
            tmp_results.append(res)
        return tmp_results

    def set_zero_forces_for_fixed_images(self):
        """This is always done in cartesian coordinates, independent
        of the actual coord_type of the images as setting forces only
        work with cartesian forces."""
        zero_forces = np.zeros_like(self.images[0].cart_coords)
        if self.fix_first:
            self.images[0].cart_forces = zero_forces
            self.log("Zeroed forces on fixed first image.")
        if self.fix_last:
            self.images[-1].cart_forces = zero_forces
            self.log("Zeroed forces on fixed last image.")

    def get_tangent(
        self, i, kind="upwinding", lanczos_guess=None, disable_lanczos=False
    ):
        """[1] Equations (8) - (11)"""

        # Converge to lowest curvature mode at the climbing image.
        # In the current implementation the given kind may be overwritten when
        # Lanczos iterations are enabled and there are climbing images. By
        # setting 'disable_lanczos=True' the provided kind is never overwritten.
        if (
            not disable_lanczos
            and self.started_climbing_lanczos
            # and (i in self.get_climbing_indices())
            and (i == self.get_hei_index())
        ):
            kind = "lanczos"

        tangent_kinds = ("upwinding", "simple", "bisect", "lanczos")
        assert kind in tangent_kinds, "Invalid kind! Valid kinds are: {tangent_kinds}"
        prev_index = max(i - 1, 0)
        next_index = min(i + 1, len(self.images) - 1)

        prev_image = self.images[prev_index]
        ith_image = self.images[i]
        next_image = self.images[next_index]

        # If (i == 0) or (i == len(self.images)-1) then one
        # of this tangents is zero.
        tangent_plus = next_image - ith_image
        tangent_minus = ith_image - prev_image

        # Handle first and last image
        if i == 0:
            return tangent_plus / np.linalg.norm(tangent_plus)
        elif i == (len(self.images) - 1):
            return tangent_minus / np.linalg.norm(tangent_minus)

        # [1], Eq. (1)
        if kind == "simple":
            tangent = next_image - prev_image
        # [1], Eq. (2)
        elif kind == "bisect":
            first_term = tangent_minus / np.linalg.norm(tangent_minus)
            sec_term = tangent_plus / np.linalg.norm(tangent_plus)
            tangent = first_term + sec_term
        # Upwinding tangent from [1] Eq. (8) and so on
        elif kind == "upwinding":
            prev_energy = prev_image.energy
            ith_energy = ith_image.energy
            next_energy = next_image.energy

            next_energy_diff = abs(next_energy - ith_energy)
            prev_energy_diff = abs(prev_energy - ith_energy)
            delta_energy_max = max(next_energy_diff, prev_energy_diff)
            delta_energy_min = min(next_energy_diff, prev_energy_diff)

            # Uphill
            if next_energy > ith_energy > prev_energy:
                tangent = tangent_plus
            # Downhill
            elif next_energy < ith_energy < prev_energy:
                tangent = tangent_minus
            # Minimum or Maximum
            else:
                if next_energy >= prev_energy:
                    tangent = (
                        tangent_plus * delta_energy_max
                        + tangent_minus * delta_energy_min
                    )
                # next_energy < prev_energy
                else:
                    tangent = (
                        tangent_plus * delta_energy_min
                        + tangent_minus * delta_energy_max
                    )
        elif kind == "lanczos":
            # Calculating a lanczos tangent is costly, so we store the
            # tangent in a dictionary. The current coordinates are
            # stringified with precision=4 and then hashed. The tangent
            # is stored/looked up with this hash.
            cur_hash = hash_arr(ith_image.coords, precision=4)
            try:
                tangent = self.lanczos_tangents[cur_hash]
                self.log(
                    "Returning previously calculated Lanczos tangent with "
                    f"hash={cur_hash}"
                )
            except KeyError:
                # Try to use previous Lanczos tangent
                guess = lanczos_guess
                if (guess is None) and (self.prev_lanczos_hash is not None):
                    guess = self.lanczos_tangents[self.prev_lanczos_hash]
                    self.log(
                        f"Using tangent with hash={self.prev_lanczos_hash} "
                        "as initial guess for Lanczos algorithm."
                    )
                w_min, tangent = geom_lanczos(
                    ith_image, guess=guess, logger=self.logger
                )
                self.lanczos_tangents[cur_hash] = tangent
                # Update hash
                self.prev_lanczos_hash = cur_hash

        tangent /= np.linalg.norm(tangent)
        return tangent

    def get_tangents(self):
        return np.array([self.get_tangent(i) for i in range(len(self.images))])

    def as_xyz(self, comments=None):
        return "\n".join([image.as_xyz() for image in self.images])

    def get_dask_client(self):
        return Client(self.scheduler)

    def get_hei_index(self, energies=None):
        """Return index of highest energy image."""
        if energies is None:
            energies = [image.energy for image in self.images]
        return np.argmax(energies)

    def prepare_opt_cycle(self, last_coords, last_energies, last_forces):
        """Implements additional logic in preparation of the next
        optimization cycle.

        Should be called by the optimizer at the beginning of a new
        optimization cycle. Can be used to implement additional logic
        as needed for AdaptiveNEB etc.
        """
        self.coords_list.append(last_coords)
        self.forces_list.append(last_forces)

        # Return False if we don't want to climb or are already
        # climbing.
        already_climbing = self.started_climbing
        if self.climb and not already_climbing:
            self.started_climbing = self.check_for_climbing_start(self.climb_rms)
            if self.started_climbing:
                msg = "Will use climbing image(s) in next cycle."
                self.log(msg)
                print(msg)
        # Determine climbing index/indices if not set, but requested.
        if already_climbing and self.climb_fixed and (self.fixed_climb_indices is None):
            self.fixed_climb_indices = self.get_climbing_indices()

        already_climbing_lanczos = self.started_climbing_lanczos
        if (
            self.climb_lanczos
            and self.started_climbing
            and not already_climbing_lanczos
        ):
            self.started_climbing_lanczos = self.check_for_climbing_start(
                self.climb_lanczos_rms
            )
            if self.started_climbing_lanczos:
                msg = "Will use Lanczos algorithm for HEI tangent in next cycle."
                self.log(msg)
                print(msg)

        return not already_climbing and self.started_climbing

    def rms(self, arr):
        """Root mean square

        Returns the root mean square of the given array.

        Parameters
        ----------
        arr : iterable of numbers

        Returns
        -------
        rms : float
            Root mean square of the given array.
        """
        return np.sqrt(np.mean(np.square(arr)))

    def check_for_climbing_start(self, ref_rms):
        # Only initiate climbing on a sufficiently converged MEP.
        # This can be determined from a supplied threshold for the
        # RMS force (rms_force) or from a multiple of the
        # RMS force convergence threshold (rms_multiple, default).
        rms_forces = self.rms(self.forces_list[-1])
        # Only start climbing when the COS is fully grown. This
        # attribute may not be defined in all subclasses, so it
        # defaults to True here.
        try:
            fully_grown = self.fully_grown
        except AttributeError:
            fully_grown = True
        start_climbing = (rms_forces <= ref_rms) and fully_grown
        return start_climbing

    def get_climbing_indices(self):
        # Index of the highest energy image (HEI)
        hei_index = self.get_hei_index()

        move_inds = self.moving_indices
        # Don't climb if not yet enabled or requested.
        if not (self.climb and self.started_climbing):
            climb_indices = tuple()
        elif self.fixed_climb_indices is not None:
            climb_indices = self.fixed_climb_indices
            _ = "index" if len(climb_indices) == 1 else "indices"
            self.log(f"Returning fixed climbing {_}.")
        # Do one image climbing (C1) neb if explicitly requested or
        # the HEI is the first or last item in moving_indices.
        elif self.climb == "one" or ((hei_index == 1) or (hei_index == move_inds[-1])):
            climb_indices = (hei_index,)
        # We can do two climbing (C2) neb if the highest energy image (HEI)
        # is in moving_indices but not the first or last item in this list.
        # elif self.climb != "one" and hei_index in move_inds[1:-1]:
        elif hei_index in move_inds[1:-1]:
            climb_indices = (hei_index - 1, hei_index + 1)
            # climb_indices = (hei_index,)
        # Don't climb when the HEI is the first or last image of the whole
        # NEB.
        else:
            climb_indices = tuple()
            self.log("Want to climb but can't. HEI is first or last image!")
        # self.log(f"Climbing indices: {climb_indices}")
        return climb_indices

    def get_climbing_forces(self, ind):
        climbing_image = self.images[ind]
        ci_forces = climbing_image.forces
        tangent = self.get_tangent(ind)
        climbing_forces = ci_forces - 2 * ci_forces.dot(tangent) * tangent

        return climbing_forces, climbing_image.energy

    def set_climbing_forces(self, forces):
        # Avoids calling the other methods with their logging output etc.
        if not self.started_climbing:
            return forces

        for i in self.get_climbing_indices():
            climb_forces, climb_en = self.get_climbing_forces(i)
            forces[i] = climb_forces
            norm = np.linalg.norm(climb_forces)
            self.log(
                f"Climbing with image {i}, E = {climb_en:.6f} au, "
                f"norm(forces)={norm:.6f}"
            )
        return forces

    def get_splined_hei(self):
        self.log("Splining HEI")
        # Interpolate energies
        cart_coords = align_coords([image.cart_coords for image in self.images])
        coord_diffs = get_coords_diffs(cart_coords)
        self.log(f"\tCoordinate differences: {coord_diffs}")
        energies = np.array(self.energy)
        energies_spline = interp1d(coord_diffs, energies, kind="cubic")
        x_fine = np.linspace(0, 1, 500)
        energies_fine = energies_spline(x_fine)
        # Determine index that yields the highest energy
        hei_ind = energies_fine.argmax()
        hei_x = x_fine[hei_ind]
        self.log(f"Found splined HEI at x={hei_x:.4f}")
        hei_frac_index = hei_x * (len(self.images) - 1)
        hei_energy = energies_fine[hei_ind]

        reshaped = cart_coords.reshape(-1, self.cart_coords_length)
        # To use splprep we have to transpose the coords.
        transp_coords = reshaped.transpose()
        tcks, us = zip(
            *[
                splprep(transp_coords[i : i + 9], s=0, k=3, u=coord_diffs)
                for i in range(0, len(transp_coords), 9)
            ]
        )

        # Reparametrize mesh
        hei_coords = np.vstack(
            [
                # WTF, Black? This looks horrible.
                splev(
                    [
                        hei_x,
                    ],
                    tck,
                )
                for tck in tcks
            ]
        )
        hei_coords = hei_coords.flatten()

        # Actually it looks like that splined tangents are really bad approximations
        # to the actual imaginary mode. The Cartesian upwinding tangent is usually
        # much much better. In 'run_tsopt_from_cos' we actually mix two "normal" tangents
        # to obtain the HEI tangent.
        hei_tangent = np.vstack(
            [
                # WTF, Black? This looks horrible.
                splev(
                    [
                        hei_x,
                    ],
                    tck,
                    der=1,
                )
                for tck in tcks
            ]
        ).T
        hei_tangent = hei_tangent.flatten()
        hei_tangent /= np.linalg.norm(hei_tangent)
        return hei_coords, hei_energy, hei_tangent, hei_frac_index

    def get_image_calc_counter_sum(self):
        return sum([image.calculator.calc_counter for image in self.images])

    def describe(self):
        imgs = self.images
        img = imgs[0]
        return f"ChainOfStates, {len(imgs)} images, ({img.sum_formula}, {len(img.atoms)} atoms) per image"

    def __str__(self):
        return self.__class__.__name__



import logging

__all__ = [
    "DampedVelocityVerlet",
    "Euler",
    "EulerPC",
    "LQA",
    "GonzalezSchlegel",
    "IMKMod",
    "ModeKill",
    "RK4",
]

from pysisyphus.irc.DampedVelocityVerlet import DampedVelocityVerlet
from pysisyphus.irc.Euler import Euler
from pysisyphus.irc.EulerPC import EulerPC
from pysisyphus.irc.GonzalezSchlegel import GonzalezSchlegel
from pysisyphus.irc.IMKMod import IMKMod
from pysisyphus.irc.LQA import LQA
from pysisyphus.irc.ModeKill import ModeKill
from pysisyphus.irc.RK4 import RK4

logger = logging.getLogger("irc")
logger.setLevel(logging.DEBUG)
# delay = True prevents creation of empty logfiles
handler = logging.FileHandler("irc.log", mode="w", delay=True)
# fmt_str = "%(levelname)s - %(message)s"
fmt_str = "%(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)



import dataclasses
from typing import Optional

try:
    from dftd4.interface import DampingParam, DispersionModel

    HAS_DFTD4 = True
except ModuleNotFoundError:
    HAS_DFTD4 = False
    # Dummy definition, so pysisyphus does not crash if another module tries to
    # import DFT4.
    DFTD4 = None

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.elem_data import nuc_charges_for_atoms


@dataclasses.dataclass
class D4Params:
    s6: float
    s8: float
    s9: float
    a1: float
    a2: float


OWN_METHODS = {
    # ORCA 5.0.3 ... your r²scan-3c bug ... why are you doing this to me? Why?
    "r2scan-3c": D4Params(s6=1.0, s8=0.0, s9=2.0, a1=0.42, a2=5.65),
}


class DFTD4(Calculator):
    def __init__(
        self,
        method: Optional[str] = None,
        damp_params: Optional[D4Params] = None,
        model_params: Optional[dict] = None,
        **kwargs
    ):
        assert HAS_DFTD4, "DFTD4 python package is not installed!"
        super().__init__(**kwargs)

        if model_params is None:
            model_params = dict()
        self.model_params = model_params

        if method in OWN_METHODS:
            damp_params = OWN_METHODS[method]

        if damp_params is not None:
            self.damp_param = DampingParam(**dataclasses.asdict(damp_params))
        # if method is not None:
        elif method is not None:
            self.method = method
            self.damp_param = DampingParam(method=method)
        else:
            raise Exception("Please provide either 'method' or 'params'!")

    def get_model(self, atoms, coords):
        atomic_numbers = nuc_charges_for_atoms(atoms)
        model = DispersionModel(
            atomic_numbers, coords.reshape(-1, 3), **self.model_params
        )
        return model

    def get_dispersion(self, atoms, coords, grad):
        model = self.get_model(atoms, coords)
        res = model.get_dispersion(self.damp_param, grad=grad)
        results = {
            "energy": float(res.get("energy")),
        }
        if grad:
            results["forces"] = -res.get("gradient").flatten()
        return results

    def get_energy(self, atoms, coords, **prepare_kwargs):
        return self.get_dispersion(atoms, coords, grad=False)

    def get_forces(self, atoms, coords, **prepare_kwargs):
        return self.get_dispersion(atoms, coords, grad=True)

    # This calculator relies on a finite differences numerical Hessian, as provided
    # by the Calculator parent class.
